from datetime import datetime

from django.template import Template
from django.test import TestCase, SimpleTestCase
from django_dynamic_fixture import N, G
from entity.models import Entity, EntityKind, EntityRelationship
from freezegun import freeze_time
from mock import patch, call, Mock
from six import text_type

from entity_event.models import (
    Medium, Source, SourceGroup, Unsubscription, Subscription, Event, EventActor, EventSeen,
    RenderingStyle, ContextRenderer, _unseen_event_ids
)
from entity_event.tests.models import TestFKModel


class EventRenderTest(TestCase):
    """
    Does an entire integration test for rendering events relative to mediums.
    """
    def test_one_context_renderer_one_medium(self):
        rg = G(RenderingStyle)
        s = G(Source)
        G(
            ContextRenderer, source=s, rendering_style=rg, text_template_path='test_template.txt',
            html_template_path='test_template.html', context_hints={
                'fk_model': {
                    'app_name': 'tests',
                    'model_name': 'TestFKModel',
                }
            })
        m = G(Medium, rendering_style=rg)

        fkm = G(TestFKModel, value=100)
        G(Event, source=s, context={'fk_model': fkm.id})

        events = Event.objects.all().load_contexts_and_renderers(m)
        txt, html = events[0].render(m)

        self.assertEquals(txt, 'Test text template with value 100')
        self.assertEquals(html, 'Test html template with value 100')

    def test_wo_fetching_contexts(self):
        rg = G(RenderingStyle)
        s = G(Source)
        G(
            ContextRenderer, source=s, rendering_style=rg, text_template_path='test_template.txt',
            html_template_path='test_template.html', context_hints={
                'fk_model': {
                    'app_name': 'tests',
                    'model_name': 'TestFKModel',
                }
            })
        m = G(Medium, rendering_style=rg)

        fkm = G(TestFKModel, value=100)
        e = G(Event, source=s, context={'fk_model': fkm.id})

        with self.assertRaises(RuntimeError):
            e.render(m)


class EventManagerCreateEventTest(TestCase):
    def test_create_event_no_actors(self):
        source = G(Source)
        e = Event.objects.create_event(context={'hi': 'hi'}, source=source)
        self.assertEqual(e.source, source)
        self.assertEqual(e.context, {'hi': 'hi'})
        self.assertEqual(e.uuid, '')
        self.assertFalse(EventActor.objects.exists())

    def test_create_event_multiple_actor_pks(self):
        source = G(Source)
        actors = [G(Entity), G(Entity)]
        e = Event.objects.create_event(context={'hi': 'hi'}, source=source, actors=[a.id for a in actors], uuid='hi')
        self.assertEqual(e.source, source)
        self.assertEqual(e.context, {'hi': 'hi'})
        self.assertEqual(e.uuid, 'hi')
        self.assertEqual(
            set(EventActor.objects.filter(event=e).values_list('entity', flat=True)), set([a.id for a in actors]))

    def test_create_event_multiple_actors(self):
        source = G(Source)
        actors = [G(Entity), G(Entity)]
        e = Event.objects.create_event(context={'hi': 'hi'}, source=source, actors=actors, uuid='hi')
        self.assertEqual(e.source, source)
        self.assertEqual(e.context, {'hi': 'hi'})
        self.assertEqual(e.uuid, 'hi')
        self.assertEqual(
            set(EventActor.objects.filter(event=e).values_list('entity', flat=True)), set([a.id for a in actors]))

    def test_ignore_duplicates_w_uuid_doesnt_already_exist(self):
        source = G(Source)
        e = Event.objects.create_event(context={'hi': 'hi'}, source=source, uuid='1', ignore_duplicates=True)
        self.assertIsNotNone(e)

    def test_ignore_duplicates_w_uuid_already_exist(self):
        source = G(Source)
        Event.objects.create_event(context={'hi': 'hi'}, source=source, uuid='1', ignore_duplicates=True)
        e = Event.objects.create_event(context={'hi': 'hi'}, source=source, uuid='1', ignore_duplicates=True)
        self.assertIsNone(e)

    def test_ignore_duplicates_wo_uuid_already_exist(self):
        source = G(Source)
        Event.objects.create_event(context={'hi': 'hi'}, source=source, ignore_duplicates=True)
        e = Event.objects.create_event(context={'hi': 'hi'}, source=source, ignore_duplicates=True)
        self.assertIsNone(e)


class EventManagerQuerySetTest(TestCase):
    def test_mark_seen(self):
        event = G(Event, context={})
        medium = G(Medium)
        Event.objects.mark_seen(medium)
        self.assertEqual(EventSeen.objects.count(), 1)
        self.assertTrue(EventSeen.objects.filter(event=event, medium=medium).exists())

    @patch('entity_event.context_loader.load_contexts_and_renderers', spec_set=True)
    def test_load_contexts_and_renderers(self, mock_load_contexts_and_renderers):
        e = G(Event, context={})
        medium = G(Medium)
        Event.objects.load_contexts_and_renderers(medium)
        self.assertEquals(mock_load_contexts_and_renderers.call_count, 1)
        self.assertEquals(list(mock_load_contexts_and_renderers.call_args_list[0][0][0]), [e])
        self.assertEquals(mock_load_contexts_and_renderers.call_args_list[0][0][1], [medium])


class MediumEventsInterfacesTest(TestCase):
    def setUp(self):
        # Set Up Entities and Relationships
        everyone_kind = G(EntityKind, name='all', display_name='all')
        group_kind = G(EntityKind, name='group', display_name='Group')
        self.person_kind = G(EntityKind, name='person', display_name='Person')

        self.p1 = G(Entity, entity_kind=self.person_kind, display_name='p1')
        self.p2 = G(Entity, entity_kind=self.person_kind, display_name='p2')
        self.p3 = G(Entity, entity_kind=self.person_kind, display_name='p3')
        p4 = G(Entity, entity_kind=self.person_kind, display_name='p4')

        g1 = G(Entity, entity_kind=group_kind)
        g2 = G(Entity, entity_kind=group_kind)

        everyone = G(Entity, entity_kind=everyone_kind)

        for sup, sub in [(g1, self.p1), (g1, self.p2), (g2, self.p3), (g2, p4)]:
            G(EntityRelationship, super_entity=sup, sub_entity=sub)
        for p in [self.p1, self.p2, self.p3, p4]:
            G(EntityRelationship, super_entity=everyone, sub_entity=p)

        # Set up Mediums, Sources, Subscriptions, Events
        self.medium_x = G(Medium, name='x', display_name='x')
        self.medium_y = G(Medium, name='y', display_name='y')
        self.medium_z = G(Medium, name='z', display_name='z')
        self.source_a = G(Source, name='a', display_name='a')
        self.source_b = G(Source, name='b', display_name='b')
        self.source_c = G(Source, name='c', display_name='c')

        e1 = G(Event, source=self.source_a, context={})
        G(Event, source=self.source_a, context={})
        e3 = G(Event, source=self.source_b, context={})
        e4 = G(Event, source=self.source_c, context={})

        G(EventActor, event=e1, entity=self.p1)
        G(EventActor, event=e3, entity=self.p2)
        G(EventActor, event=e4, entity=self.p2)
        G(EventActor, event=e4, entity=self.p3)

        G(Subscription, source=self.source_a, medium=self.medium_x, only_following=False,
          entity=everyone, sub_entity_kind=self.person_kind)
        G(Subscription, source=self.source_a, medium=self.medium_y, only_following=True,
          entity=everyone, sub_entity_kind=self.person_kind)
        G(Subscription, source=self.source_c, medium=self.medium_z, only_following=True,
          entity=g1, sub_entity_kind=self.person_kind)

    def test_events_basic(self):
        events = self.medium_x.events()
        self.assertEqual(events.count(), 2)

    def test_events_only_following(self):
        events = self.medium_y.events()
        self.assertEqual(events.count(), 1)

    def test_entity_events_basic(self):
        events = self.medium_x.entity_events(entity=self.p1)
        self.assertEqual(len(events), 2)

    def test_entity_events_basic_mark_seen(self):
        events = self.medium_x.entity_events(entity=self.p1, seen=False, mark_seen=True)
        self.assertEqual(len(events), 2)
        # All unseen events should have been marked as seen, even if they werent related
        # to the entity
        self.assertEqual(len(EventSeen.objects.all()), 4)

    def test_entity_events_basic_unsubscribed(self):
        G(Unsubscription, entity=self.p1, source=self.source_a, medium=self.medium_x)
        G(Event, source=self.source_b, context={})
        G(Subscription, source=self.source_b, medium=self.medium_x, only_following=False,
          entity=self.p1, sub_entity_kind=None)
        events = self.medium_x.entity_events(entity=self.p1)
        self.assertEqual(len(events), 2)
        for event in events:
            self.assertEqual(event.source, self.source_b)

    def test_entity_events_only_following(self):
        events = self.medium_z.entity_events(entity=self.p2)
        self.assertEqual(len(events), 1)

    def test_entity_targets_basic(self):
        events_targets = self.medium_x.events_targets()
        self.assertEqual(len(events_targets), 2)

    def test_entity_targets_target_count(self):
        events_targets = self.medium_x.events_targets(entity_kind=self.person_kind)
        self.assertEqual(len(events_targets[0][1]), 4)

    def test_entity_targets_only_following(self):
        events_targets = self.medium_z.events_targets(entity_kind=self.person_kind)
        self.assertEqual(len(events_targets[0][1]), 1)


class MediumRenderTest(SimpleTestCase):
    @patch('entity_event.context_loader.load_contexts_and_renderers', spec_set=True)
    def test_render(self, mock_load_contexts_and_renderers):
        medium = N(Medium, persist_dependencies=False)
        e1 = Mock(render=Mock(return_value=('e1.txt', 'e1.html')))
        e2 = Mock(render=Mock(return_value=('e2.txt', 'e2.html')))

        events = [e1, e2]
        res = medium.render(events)

        mock_load_contexts_and_renderers.assert_called_once_with(events, [medium])
        self.assertEquals(res, {
            e1: ('e1.txt', 'e1.html'),
            e2: ('e2.txt', 'e2.html'),
        })
        e1.render.assert_called_once_with(medium)
        e2.render.assert_called_once_with(medium)


class MediumSubsetSubscriptionsTest(TestCase):
    def setUp(self):
        person = G(EntityKind, name='person', display_name='Person')
        self.super_e = G(Entity)
        self.sub_e = G(Entity, entity_kind=person)
        random = G(Entity)
        G(EntityRelationship, super_entity=self.super_e, sub_entity=self.sub_e)

        self.medium = G(Medium)
        self.group_sub = G(Subscription, entity=self.super_e, sub_entity_kind=person)
        self.indiv_sub = G(Subscription, entity=self.sub_e, sub_entity_kind=None)
        self.random_sub = G(Subscription, entity=random)

    def test_no_entity(self):
        all_subs = Subscription.objects.all()
        subs = self.medium.subset_subscriptions(all_subs)
        self.assertEqual(subs, all_subs)

    def test_sub_entity(self):
        all_subs = Subscription.objects.all()
        subs = self.medium.subset_subscriptions(all_subs, self.sub_e)
        self.assertEqual(subs.count(), 2)

    def test_super_not_included(self):
        all_subs = Subscription.objects.all()
        subs = self.medium.subset_subscriptions(all_subs, self.super_e)
        self.assertEqual(subs.count(), 0)


class MediumGetFilteredEventsTest(TestCase):
    def setUp(self):
        self.medium = G(Medium)

    def test_get_unseen_events_some_seen_some_not(self):
        seen_e = G(Event, context={})
        G(EventSeen, event=seen_e, medium=self.medium)
        unseen_e = G(Event, context={})

        events = self.medium.get_filtered_events(seen=False)
        self.assertEquals(list(events), [unseen_e])

    def test_get_unseen_events_some_seen_from_other_mediums(self):
        seen_from_other_medium_e = G(Event, context={})
        G(EventSeen, event=seen_from_other_medium_e)
        unseen_e = G(Event, context={})

        events = self.medium.get_filtered_events(seen=False)
        self.assertEquals(set(events), set([unseen_e, seen_from_other_medium_e]))


class MediumGetEventFiltersTest(TestCase):
    def setUp(self):
        self.medium = G(Medium)
        with freeze_time('2014-01-15'):
            e1 = G(Event, context={})
            G(Event, context={}, time_expires=datetime(5000, 1, 1))
        with freeze_time('2014-01-17'):
            G(Event, context={}), G(Event, context={})
            G(Event, context={}, time_expires=datetime(2014, 1, 17))
        G(EventSeen, event=e1, medium=self.medium)

        self.actor = G(Entity)
        G(EventActor, event=e1, entity=self.actor)

    def test_start_time(self):
        filters = self.medium.get_filtered_events_queries(datetime(2014, 1, 16), None, None, True, None)
        events = Event.objects.filter(*filters)
        self.assertEqual(events.count(), 3)

    def test_end_time(self):
        filters = self.medium.get_filtered_events_queries(None, datetime(2014, 1, 16), None, True, None)
        events = Event.objects.filter(*filters)
        self.assertEqual(events.count(), 2)

    def test_is_seen(self):
        filters = self.medium.get_filtered_events_queries(None, None, True, True, None)
        events = Event.objects.filter(*filters)
        self.assertEqual(events.count(), 1)

    def test_is_not_seen(self):
        filters = self.medium.get_filtered_events_queries(None, None, False, True, None)
        events = Event.objects.filter(*filters)
        self.assertEqual(events.count(), 4)

    def test_include_expires(self):
        filters = self.medium.get_filtered_events_queries(None, None, None, True, None)
        events = Event.objects.filter(*filters)
        self.assertEqual(events.count(), 5)

    def test_dont_include_expires(self):
        filters = self.medium.get_filtered_events_queries(None, None, None, False, None)
        events = Event.objects.filter(*filters)
        self.assertEqual(events.count(), 4)

    def test_actor(self):
        filters = self.medium.get_filtered_events_queries(None, None, None, True, self.actor)
        events = Event.objects.filter(*filters)
        self.assertEqual(events.count(), 1)


class MediumFollowedByTest(TestCase):
    def setUp(self):
        self.medium = N(Medium, persist_dependencies=False)
        self.superentity = G(Entity)
        self.sub1, self.sub2 = G(Entity), G(Entity)
        G(EntityRelationship, super_entity=self.superentity, sub_entity=self.sub1)
        G(EntityRelationship, super_entity=self.superentity, sub_entity=self.sub2)

    def test_self_in(self):
        followers = self.medium.followed_by(self.sub1)
        super_entity_in = followers.filter(id=self.sub1.id).exists()
        self.assertTrue(super_entity_in)

    def test_super_entities_in(self):
        followers = self.medium.followed_by(self.sub1)
        sub_entity_in = followers.filter(id=self.superentity.id).exists()
        self.assertTrue(sub_entity_in)

    def test_others_not_in(self):
        followers = self.medium.followed_by(self.sub1)
        random_entity_in = followers.filter(id=self.sub2.id).exists()
        self.assertFalse(random_entity_in)

    def test_multiple_inputs_list(self):
        followers = self.medium.followed_by([self.sub1.id, self.sub2.id])
        self.assertEqual(followers.count(), 3)

    def test_multiple_inputs_qs(self):
        entities = Entity.objects.filter(id__in=[self.sub1.id, self.sub2.id])
        followers = self.medium.followed_by(entities)
        self.assertEqual(followers.count(), 3)


class MediumFollowersOfTest(TestCase):
    def setUp(self):
        self.medium = N(Medium, persist_dependencies=False)
        self.superentity = G(Entity)
        self.sub1, self.sub2 = G(Entity), G(Entity)
        self.random_entity = G(Entity)
        G(EntityRelationship, super_entity=self.superentity, sub_entity=self.sub1)
        G(EntityRelationship, super_entity=self.superentity, sub_entity=self.sub2)

    def test_self_in(self):
        followers = self.medium.followers_of(self.superentity)
        super_entity_in = followers.filter(id=self.superentity.id).exists()
        self.assertTrue(super_entity_in)

    def test_sub_entities_in(self):
        followers = self.medium.followers_of(self.superentity)
        sub_entity_in = followers.filter(id=self.sub1.id).exists()
        self.assertTrue(sub_entity_in)

    def test_others_not_in(self):
        followers = self.medium.followers_of(self.superentity)
        random_entity_in = followers.filter(id=self.random_entity.id).exists()
        self.assertFalse(random_entity_in)

    def test_multiple_inputs_list(self):
        followers = self.medium.followers_of([self.sub1.id, self.sub2.id])
        self.assertEqual(followers.count(), 2)

    def test_multiple_inputs_qs(self):
        entities = Entity.objects.filter(id__in=[self.sub1.id, self.sub2.id])
        followers = self.medium.followers_of(entities)
        self.assertEqual(followers.count(), 2)


class SubscriptionSubscribedEntitiesTest(TestCase):
    def setUp(self):
        person_kind = G(EntityKind, name='person', display_name='person')
        superentity = G(Entity)
        sub1, sub2 = G(Entity, entity_kind=person_kind), G(Entity, entity_kind=person_kind)
        G(EntityRelationship, super_entity=superentity, sub_entity=sub1)
        G(EntityRelationship, super_entity=superentity, sub_entity=sub2)

        self.group_sub = N(Subscription, entity=superentity, sub_entity_kind=person_kind, persist_dependencies=False)
        self.indiv_sub = N(Subscription, entity=superentity, sub_entity_kind=None, persist_dependencies=False)

    def test_both_branches_return_queryset(self):
        group_qs = self.group_sub.subscribed_entities()
        indiv_qs = self.indiv_sub.subscribed_entities()
        self.assertEqual(type(group_qs), type(indiv_qs))

    def test_length_group(self):
        group_qs = self.group_sub.subscribed_entities()
        self.assertEqual(group_qs.count(), 2)

    def test_length_indiv(self):
        indiv_qs = self.indiv_sub.subscribed_entities()
        self.assertEqual(indiv_qs.count(), 1)


class ContextRendererRenderTextOrHtmlTemplateTest(SimpleTestCase):
    @patch('entity_event.models.render_to_string', spec_set=True)
    def test_w_html_template_path(self, mock_render_to_string):
        cr = N(ContextRenderer, html_template_path='html_path', persist_dependencies=False)
        c = {'context': 'context'}
        cr.render_text_or_html_template(c, is_text=False)
        mock_render_to_string.assert_called_once_with('html_path', c)

    @patch('entity_event.models.render_to_string', spec_set=True)
    def test_w_text_template_path(self, mock_render_to_string):
        cr = N(ContextRenderer, text_template_path='text_path', persist_dependencies=False)
        c = {'context': 'context'}
        cr.render_text_or_html_template(c, is_text=True)
        mock_render_to_string.assert_called_once_with('text_path', c)

    @patch.object(Template, '__init__', spec_set=True, return_value=None)
    @patch.object(Template, 'render', spec_set=True)
    def test_w_html_template(self, mock_render, mock_init):
        cr = N(ContextRenderer, html_template='html_template', persist_dependencies=False)
        c = {'context': 'context'}
        cr.render_text_or_html_template(c, is_text=False)
        self.assertEqual(mock_render.call_count, 1)
        mock_init.assert_called_once_with('html_template')

    @patch.object(Template, '__init__', spec_set=True, return_value=None)
    @patch.object(Template, 'render', spec_set=True)
    def test_w_text_template(self, mock_render, mock_init):
        cr = N(ContextRenderer, text_template='text_template', persist_dependencies=False)
        c = {'context': 'context'}
        cr.render_text_or_html_template(c, is_text=True)
        self.assertEqual(mock_render.call_count, 1)
        mock_init.assert_called_once_with('text_template')

    def test_w_no_templates_text(self):
        cr = N(ContextRenderer, persist_dependencies=False)
        c = {'context': 'context'}
        self.assertEqual(cr.render_text_or_html_template(c, is_text=True), '')

    def test_w_no_templates_html(self):
        cr = N(ContextRenderer, persist_dependencies=False)
        c = {'context': 'context'}
        self.assertEqual(cr.render_text_or_html_template(c, is_text=False), '')


class ContextRendererRenderContextToTextHtmlTemplates(SimpleTestCase):
    @patch.object(ContextRenderer, 'render_text_or_html_template', spec_set=True)
    def test_render_context_to_text_html_templates(self, mock_render_text_or_html_template):
        c = {'context': 'context'}
        r = ContextRenderer().render_context_to_text_html_templates(c)
        self.assertEqual(
            r, (
                mock_render_text_or_html_template.return_value.strip(),
                mock_render_text_or_html_template.return_value.strip()
            ))
        self.assertEqual(
            mock_render_text_or_html_template.call_args_list, [call(c, is_text=True), call(c, is_text=False)])


class UnseenEventIdsTest(TestCase):
    def test_filters_seen(self):
        m = G(Medium)
        e1 = G(Event, context={})
        e2 = G(Event, context={})
        Event.objects.filter(id=e2.id).mark_seen(m)
        unseen_ids = _unseen_event_ids(m)
        self.assertEqual(unseen_ids, [e1.id])

    def test_multiple_mediums(self):
        m1 = G(Medium)
        m2 = G(Medium)
        e1 = G(Event, context={})
        e2 = G(Event, context={})
        e3 = G(Event, context={})
        e4 = G(Event, context={})
        Event.objects.filter(id=e2.id).mark_seen(m1)
        Event.objects.filter(id=e3.id).mark_seen(m2)
        unseen_ids = _unseen_event_ids(m1)
        self.assertEqual(set(unseen_ids), set([e1.id, e3.id, e4.id]))


class UnicodeTest(SimpleTestCase):
    def setUp(self):
        self.rendering_style = N(RenderingStyle, display_name='Test Render Group', persist_dependencies=False)
        self.context_renderer = N(ContextRenderer, name='Test Context Renderer', persist_dependencies=False)
        self.medium = N(Medium, display_name='Test Medium', persist_dependencies=False)
        self.source = N(Source, display_name='Test Source', persist_dependencies=False)
        self.source_group = N(SourceGroup, display_name='Test Source Group', persist_dependencies=False)
        self.entity = N(Entity, display_name='Test Entity', persist_dependencies=False)
        self.unsubscription = N(
            Unsubscription, entity=self.entity, medium=self.medium, source=self.source, persist_dependencies=False)
        self.subscription = N(
            Subscription, entity=self.entity, source=self.source, medium=self.medium, persist_dependencies=False)
        self.event = N(Event, source=self.source, context={}, id=1, persist_dependencies=False)
        self.event_actor = N(EventActor, event=self.event, entity=self.entity, persist_dependencies=False)
        self.event_seen = N(
            EventSeen, event=self.event, medium=self.medium, time_seen=datetime(2014, 1, 2), persist_dependencies=False)

    def test_RenderingStyle_formats(self):
        s = text_type(self.rendering_style)
        self.assertEquals(s, 'Test Render Group')

    def test_contextrenderer_formats(self):
        s = text_type(self.context_renderer)
        self.assertEquals(s, 'Test Context Renderer')

    def test_medium_formats(self):
        s = text_type(self.medium)
        self.assertEqual(s, 'Test Medium')

    def test_source_formats(self):
        s = text_type(self.source)
        self.assertEqual(s, 'Test Source')

    def test_sourcegroup_formats(self):
        s = text_type(self.source_group)
        self.assertEqual(s, 'Test Source Group')

    def test_unsubscription_formats(self):
        s = text_type(self.unsubscription)
        self.assertEqual(s, '{0} from Test Source by Test Medium'.format(self.entity))

    def test_subscription_formats(self):
        s = text_type(self.subscription)
        self.assertEqual(s, '{0} to Test Source by Test Medium'.format(self.entity))

    def test_event_formats(self):
        s = text_type(self.event)
        self.assertTrue(s.startswith('Test Source event at 201'))

    def test_eventactor_formats(self):
        s = text_type(self.event_actor)
        self.assertEqual(s, 'Event 1 - {0}'.format(self.entity))

    def test_event_seenformats(self):
        s = text_type(self.event_seen)
        self.assertEqual(s, 'Seen on Test Medium at 2014-01-02::00:00:00')



"""
A module for loading contexts using context hints.
"""
from collections import defaultdict
import six

from django.db.models import Q
from django.db.models.loading import get_model
from manager_utils import id_dict

from entity_event.models import ContextRenderer


def get_context_hints_per_source(context_renderers):
    """
    Given a list of context renderers, return a dictionary of context hints per source.
    """
    # Merge the context render hints for each source as there can be multiple context hints for
    # sources depending on the render target. Merging them together involves combining select
    # and prefetch related hints for each context renderer
    context_hints_per_source = defaultdict(lambda: defaultdict(lambda: {
        'app_name': None,
        'model_name': None,
        'select_related': set(),
        'prefetch_related': set(),
    }))
    for cr in context_renderers:
        for key, hints in cr.context_hints.items() if cr.context_hints else []:
            for source in cr.get_sources():
                context_hints_per_source[source][key]['app_name'] = hints['app_name']
                context_hints_per_source[source][key]['model_name'] = hints['model_name']
                context_hints_per_source[source][key]['select_related'].update(hints.get('select_related', []))
                context_hints_per_source[source][key]['prefetch_related'].update(hints.get('prefetch_related', []))

    return context_hints_per_source


def get_querysets_for_context_hints(context_hints_per_source):
    """
    Given a list of context hint dictionaries, return a dictionary
    of querysets for efficient context loading. The return value
    is structured as follows:

    {
        model: queryset,
        ...
    }
    """
    model_select_relateds = defaultdict(set)
    model_prefetch_relateds = defaultdict(set)
    model_querysets = {}
    for context_hints in context_hints_per_source.values():
        for hints in context_hints.values():
            model = get_model(hints['app_name'], hints['model_name'])
            model_querysets[model] = model.objects
            model_select_relateds[model].update(hints.get('select_related', []))
            model_prefetch_relateds[model].update(hints.get('prefetch_related', []))

    # Attach select and prefetch related parameters to the querysets if needed
    for model, queryset in model_querysets.items():
        if model_select_relateds[model]:
            queryset = queryset.select_related(*model_select_relateds[model])
        if model_prefetch_relateds[model]:
            queryset = queryset.prefetch_related(*model_prefetch_relateds[model])
        model_querysets[model] = queryset

    return model_querysets


def dict_find(d, which_key):
    """
    Finds key values in a nested dictionary. Returns a tuple of the dictionary in which
    the key was found along with the value
    """
    # If the starting point is a list, iterate recursively over all values
    if isinstance(d, (list, tuple)):
        for i in d:
            for result in dict_find(i, which_key):
                yield result

    # Else, iterate over all key values of the dictionary
    elif isinstance(d, dict):
        for k, v in d.items():
            if k == which_key:
                yield d, v
            for result in dict_find(v, which_key):
                yield result


def get_model_ids_to_fetch(events, context_hints_per_source):
    """
    Obtains the ids of all models that need to be fetched. Returns a dictionary of models that
    point to sets of ids that need to be fetched. Return output is as follows:

    {
        model: [id1, id2, ...],
        ...
    }
    """
    number_types = (complex, float) + six.integer_types
    model_ids_to_fetch = defaultdict(set)

    for event in events:
        context_hints = context_hints_per_source.get(event.source, {})
        for context_key, hints in context_hints.items():
            for d, value in dict_find(event.context, context_key):
                values = value if isinstance(value, list) else [value]
                model_ids_to_fetch[get_model(hints['app_name'], hints['model_name'])].update(
                    v for v in values if isinstance(v, number_types)
                )

    return model_ids_to_fetch


def fetch_model_data(model_querysets, model_ids_to_fetch):
    """
    Given a dictionary of models to querysets and model IDs to models, fetch the IDs
    for every model and return the objects in the following structure.

    {
        model: {
            id: obj,
            ...
        },
        ...
    }
    """
    return {
        model: id_dict(model_querysets[model].filter(id__in=ids_to_fetch))
        for model, ids_to_fetch in model_ids_to_fetch.items()
    }


def load_fetched_objects_into_contexts(events, model_data, context_hints_per_source):
    """
    Given the fetched model data and the context hints for each source, go through each
    event and populate the contexts with the loaded information.
    """
    for event in events:
        context_hints = context_hints_per_source.get(event.source, {})
        for context_key, hints in context_hints.items():
            model = get_model(hints['app_name'], hints['model_name'])
            for d, value in dict_find(event.context, context_key):
                if isinstance(value, list):
                    for i, model_id in enumerate(d[context_key]):
                        d[context_key][i] = model_data[model].get(model_id)
                else:
                    d[context_key] = model_data[model].get(value)


def load_renderers_into_events(events, mediums, context_renderers):
    """
    Given the events and the context renderers, load the renderers into the event objects
    so that they may be able to call the 'render' method later on.
    """
    mediums_per_rendering_style = defaultdict(list)
    for medium in mediums:
        mediums_per_rendering_style[medium.rendering_style_id].append(medium)

    medium_renderers_per_source = defaultdict(dict)
    for renderer in context_renderers:
        for medium in mediums_per_rendering_style[renderer.rendering_style_id]:
            for source in renderer.get_sources():
                medium_renderers_per_source[source.id][medium] = renderer

    for event in events:
        for medium, renderer in medium_renderers_per_source[event.source_id].items():
            event._context_renderers[medium] = renderer


def load_contexts_and_renderers(events, mediums):
    """
    Given a list of events and mediums, load the context model data into the contexts of the events.
    """
    sources = {event.source for event in events}
    rendering_styles = {medium.rendering_style for medium in mediums}
    context_renderers = ContextRenderer.objects.filter(
        Q(source__in=sources, rendering_style__in=rendering_styles) |
        Q(source_group_id__in=[s.group_id for s in sources], rendering_style__in=rendering_styles)).select_related(
            'source').prefetch_related('source_group__source_set')

    context_hints_per_source = get_context_hints_per_source(context_renderers)
    model_querysets = get_querysets_for_context_hints(context_hints_per_source)
    model_ids_to_fetch = get_model_ids_to_fetch(events, context_hints_per_source)
    model_data = fetch_model_data(model_querysets, model_ids_to_fetch)
    load_fetched_objects_into_contexts(events, model_data, context_hints_per_source)
    load_renderers_into_events(events, mediums, context_renderers)

    return events



from django.test import TestCase
from django_dynamic_fixture import G

from entity_event.admin import AdminEventForm
from entity_event.models import Source, Event


class AdminEventFormSaveTest(TestCase):
    def setUp(self):
        source = G(Source)
        self.form_data = {
            'source': source.id,
            'text': 'test text',
        }

    def test_new_obj(self):
        form = AdminEventForm(self.form_data)
        form.is_valid()
        form.save()
        self.assertEqual(Event.objects.count(), 1)


class AdminEventFormSaveM2MTest(TestCase):
    def setUp(self):
        source = G(Source)
        self.form_data = {
            'source': source.id,
            'text': 'test text',
        }

    def test_does_nothing(self):
        form = AdminEventForm(self.form_data)
        form.save_m2m()
        self.assertEqual(Event.objects.count(), 0)



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'Medium'
        db.create_table(u'entity_event_medium', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(unique=True, max_length=64)),
            ('display_name', self.gf('django.db.models.fields.CharField')(max_length=64)),
            ('description', self.gf('django.db.models.fields.TextField')()),
        ))
        db.send_create_signal(u'entity_event', ['Medium'])

        # Adding model 'Source'
        db.create_table(u'entity_event_source', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(unique=True, max_length=64)),
            ('display_name', self.gf('django.db.models.fields.CharField')(max_length=64)),
            ('description', self.gf('django.db.models.fields.TextField')()),
            ('group', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['entity_event.SourceGroup'])),
        ))
        db.send_create_signal(u'entity_event', ['Source'])

        # Adding model 'SourceGroup'
        db.create_table(u'entity_event_sourcegroup', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(unique=True, max_length=64)),
            ('display_name', self.gf('django.db.models.fields.CharField')(max_length=64)),
            ('description', self.gf('django.db.models.fields.TextField')()),
        ))
        db.send_create_signal(u'entity_event', ['SourceGroup'])

        # Adding model 'Unsubscription'
        db.create_table(u'entity_event_unsubscription', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('entity', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['entity.Entity'])),
            ('medium', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['entity_event.Medium'])),
            ('source', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['entity_event.Source'])),
        ))
        db.send_create_signal(u'entity_event', ['Unsubscription'])

        # Adding model 'Subscription'
        db.create_table(u'entity_event_subscription', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('medium', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['entity_event.Medium'])),
            ('source', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['entity_event.Source'])),
            ('entity', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['entity.Entity'])),
            ('sub_entity_kind', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['entity.EntityKind'], null=True)),
            ('only_following', self.gf('django.db.models.fields.BooleanField')(default=True)),
        ))
        db.send_create_signal(u'entity_event', ['Subscription'])

        # Adding model 'Event'
        db.create_table(u'entity_event_event', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('source', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['entity_event.Source'])),
            ('context', self.gf('jsonfield.fields.JSONField')()),
            ('time', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
            ('time_expires', self.gf('django.db.models.fields.DateTimeField')(default=None, null=True)),
            ('uuid', self.gf('django.db.models.fields.CharField')(unique=True, max_length=128)),
        ))
        db.send_create_signal(u'entity_event', ['Event'])

        # Adding model 'EventActor'
        db.create_table(u'entity_event_eventactor', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('event', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['entity_event.Event'])),
            ('entity', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['entity.Entity'])),
        ))
        db.send_create_signal(u'entity_event', ['EventActor'])

        # Adding model 'EventSeen'
        db.create_table(u'entity_event_eventseen', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('event', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['entity_event.Event'])),
            ('medium', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['entity_event.Medium'])),
            ('time_seen', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime.utcnow)),
        ))
        db.send_create_signal(u'entity_event', ['EventSeen'])


    def backwards(self, orm):
        # Deleting model 'Medium'
        db.delete_table(u'entity_event_medium')

        # Deleting model 'Source'
        db.delete_table(u'entity_event_source')

        # Deleting model 'SourceGroup'
        db.delete_table(u'entity_event_sourcegroup')

        # Deleting model 'Unsubscription'
        db.delete_table(u'entity_event_unsubscription')

        # Deleting model 'Subscription'
        db.delete_table(u'entity_event_subscription')

        # Deleting model 'Event'
        db.delete_table(u'entity_event_event')

        # Deleting model 'EventActor'
        db.delete_table(u'entity_event_eventactor')

        # Deleting model 'EventSeen'
        db.delete_table(u'entity_event_eventseen')


    models = {
        u'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        u'entity.entity': {
            'Meta': {'unique_together': "(('entity_id', 'entity_type', 'entity_kind'),)", 'object_name': 'Entity'},
            'display_name': ('django.db.models.fields.TextField', [], {'db_index': 'True', 'blank': 'True'}),
            'entity_id': ('django.db.models.fields.IntegerField', [], {}),
            'entity_kind': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity.EntityKind']", 'on_delete': 'models.PROTECT'}),
            'entity_meta': ('jsonfield.fields.JSONField', [], {'null': 'True'}),
            'entity_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['contenttypes.ContentType']", 'on_delete': 'models.PROTECT'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True', 'db_index': 'True'})
        },
        u'entity.entitykind': {
            'Meta': {'object_name': 'EntityKind'},
            'display_name': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '256', 'db_index': 'True'})
        },
        u'entity_event.event': {
            'Meta': {'object_name': 'Event'},
            'context': ('jsonfield.fields.JSONField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'source': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity_event.Source']"}),
            'time': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'time_expires': ('django.db.models.fields.DateTimeField', [], {'default': 'None', 'null': 'True'}),
            'uuid': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '128'})
        },
        u'entity_event.eventactor': {
            'Meta': {'object_name': 'EventActor'},
            'entity': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity.Entity']"}),
            'event': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity_event.Event']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'})
        },
        u'entity_event.eventseen': {
            'Meta': {'object_name': 'EventSeen'},
            'event': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity_event.Event']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'medium': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity_event.Medium']"}),
            'time_seen': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.utcnow'})
        },
        u'entity_event.medium': {
            'Meta': {'object_name': 'Medium'},
            'description': ('django.db.models.fields.TextField', [], {}),
            'display_name': ('django.db.models.fields.CharField', [], {'max_length': '64'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '64'})
        },
        u'entity_event.source': {
            'Meta': {'object_name': 'Source'},
            'description': ('django.db.models.fields.TextField', [], {}),
            'display_name': ('django.db.models.fields.CharField', [], {'max_length': '64'}),
            'group': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity_event.SourceGroup']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '64'})
        },
        u'entity_event.sourcegroup': {
            'Meta': {'object_name': 'SourceGroup'},
            'description': ('django.db.models.fields.TextField', [], {}),
            'display_name': ('django.db.models.fields.CharField', [], {'max_length': '64'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '64'})
        },
        u'entity_event.subscription': {
            'Meta': {'object_name': 'Subscription'},
            'entity': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity.Entity']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'medium': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity_event.Medium']"}),
            'only_following': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'source': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity_event.Source']"}),
            'sub_entity_kind': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity.EntityKind']", 'null': 'True'})
        },
        u'entity_event.unsubscription': {
            'Meta': {'object_name': 'Unsubscription'},
            'entity': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity.Entity']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'medium': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity_event.Medium']"}),
            'source': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity_event.Source']"})
        }
    }

    complete_apps = ['entity_event']


from datetime import datetime
from uuid import uuid1

from django import forms
from django.contrib import admin
from django.forms.extras.widgets import SelectDateWidget

from entity_event.models import (
    AdminEvent, Event, EventActor, EventSeen, Medium, Source, SourceGroup, Subscription, Unsubscription
)


class AdminEventForm(forms.ModelForm):
    source = forms.ModelChoiceField(queryset=Source.objects.all())  # initial = Source.objects.get(name='admin')
    text = forms.CharField(widget=forms.Textarea(attrs={'rows': '3', 'cols': '60'}))
    expires_date = forms.DateField(widget=SelectDateWidget(), required=False)
    expires_time = forms.TimeField(label='Expires time (UTC 24 hr) E.g. 18:25', required=False)

    class Meta:
        model = AdminEvent
        fields = ['source', 'text', 'expires_date', 'expires_time']

    def save(self, *args, **kwargs):
        self.clean()
        expires_date = self.cleaned_data['expires_date']
        expires_time = self.cleaned_data['expires_time']
        expires_datetime = (
            datetime.combine(expires_date, expires_time) if expires_date and expires_time else datetime.max)
        context = {'text': self.cleaned_data['text']}
        event = Event.objects.create(
            source=self.cleaned_data['source'],
            context=context,
            time_expires=expires_datetime,
            uuid=uuid1().hex
        )
        return event

    def save_m2m(self, *args, **kwargs):
        pass


class AdminEventAdmin(admin.ModelAdmin):
    list_display = ('time', 'source')
    form = AdminEventForm


class EventActorInline(admin.TabularInline):
    model = EventActor


class EventSeenInline(admin.TabularInline):
    model = EventSeen


class EventAdmin(admin.ModelAdmin):
    list_display = ('time', 'source')
    inlines = [
        EventActorInline,
        EventSeenInline
    ]


class MediumAdmin(admin.ModelAdmin):
    list_display = ('display_name', 'description')


class SourceAdmin(admin.ModelAdmin):
    list_display = ('display_name', 'description')


class SourceGroupAdmin(admin.ModelAdmin):
    list_display = ('display_name', 'description')


class SubscriptionAdmin(admin.ModelAdmin):
    list_display = ('entity', 'source', 'medium', 'sub_entity_kind', 'only_following')


class UnsubscriptionAdmin(admin.ModelAdmin):
    list_display = ('entity', 'medium', 'source')


admin.site.register(AdminEvent, AdminEventAdmin)
admin.site.register(Medium, MediumAdmin)
admin.site.register(Event, EventAdmin)
admin.site.register(Source, SourceAdmin)
admin.site.register(SourceGroup, SourceGroupAdmin)
admin.site.register(Subscription, SubscriptionAdmin)
admin.site.register(Unsubscription, UnsubscriptionAdmin)



from collections import defaultdict
from datetime import datetime
from operator import or_
from six.moves import reduce

from cached_property import cached_property
from django.db import models, transaction
from django.db.models import Q
from django.db.models.query import QuerySet
from django.template.loader import render_to_string
from django.template import Context, Template
from django.utils.encoding import python_2_unicode_compatible
from entity.models import Entity, EntityKind, EntityRelationship
import jsonfield


@python_2_unicode_compatible
class Medium(models.Model):
    """A ``Medium`` is an object in the database that defines the method
    by which users will view events. The actual objects in the
    database are fairly simple, only requiring a ``name``,
    ``display_name`` and ``description``. Mediums can be created with
    ``Medium.objects.create``, using the following parameters:

    :type name: str
    :param name: A short, unique name for the medium.

    :type display_name: str
    :param display_name: A short, human readable name for the medium.
        Does not need to be unique.

    :type description: str
    :param description: A human readable description of the
        medium.

    Encoding a ``Medium`` object in the database serves two
    purposes. First, it is referenced when subscriptions are
    created. Second the ``Medium`` objects provide an entry point to
    query for events and have all the subscription logic and filtering
    taken care of for you.

    Any time a new way to display events to a user is created, a
    corresponding ``Medium`` should be created. Some examples could
    include a medium for sending email notifications, a medium for
    individual newsfeeds, or a medium for a site wide notification
    center.

    Once a medium object is created, and corresponding subscriptions
    are created, there are three methods on the medium object that can
    be used to query for events. They are ``events``,
    ``entity_events`` and ``events_targets``. The differences between
    these methods are described in their corresponding documentation.

    A medium can use a ``RenderingStyle`` to use a configured style of rendering
    with the medium. Any associated ``ContextRenderer`` models defined with
    that rendering style will be used to render events in the ``render`` method
    of the medium. This is an optional part of Entity Event's built-in
    rendering system.
    """
    name = models.CharField(max_length=64, unique=True)
    display_name = models.CharField(max_length=64)
    description = models.TextField()
    rendering_style = models.ForeignKey('RenderingStyle', null=True)

    def __str__(self):
        """Readable representation of ``Medium`` objects."""
        return self.display_name

    @transaction.atomic
    def events(self, **event_filters):
        """Return subscribed events, with basic filters.

        This method of getting events is useful when you want to
        display events for your medium, independent of what entities
        were involved in those events. For example, this method can be
        used to display a list of site-wide events that happened in the
        past 24 hours:

        .. code-block:: python

            TEMPLATE = '''
                <html><body>
                <h1> Yoursite's Events </h1>
                <ul>
                {% for event in events %}
                    <li> {{ event.context.event_text }} </li>
                {% endfor %}
                </ul>
                </body></html>
            '''

            def site_feed(request):
                site_feed_medium = Medium.objects.get(name='site_feed')
                start_time = datetime.utcnow() - timedelta(days=1)
                context = {}
                context['events'] = site_feed_medium.events(start_time=start_time)
                return HttpResponse(TEMPLATE.render(context))

        While the `events` method does not filter events based on what
        entities are involved, filtering based on the properties of the events
        themselves is supported, through the following arguments, all
        of which are optional.

        :type start_time: datetime.datetime (optional)
        :param start_time: Only return events that occurred after the
            given time. If no time is given for this argument, no
            filtering is done.

        :type end_time: datetime.datetime (optional)
        :param end_time: Only return events that occurred before the
            given time. If no time is given for this argument, no
            filtering is done

        :type seen: Boolean (optional)
        :param seen: This flag controls whether events that have
            marked as seen are included. By default, both events that
            have and have not been marked as seen are included. If
            ``True`` is given for this parameter, only events that
            have been marked as seen will be included. If ``False`` is
            given, only events that have not been marked as seen will
            be included.

        :type include_expired: Boolean (optional)
        :param include_expired: By default, events that have a
            expiration time, which has passed, are not included in the
            results. Passing in ``True`` to this argument causes
            expired events to be returned as well.

        :type actor: Entity (optional)
        :param actor: Only include events with the given entity as an
            actor.

        :type mark_seen: Boolean (optional)
        :param mark_seen: Create a side effect in the database that
            marks all the returned events as having been seen by this
            medium.

        :rtype: EventQuerySet
        :returns: A queryset of events.
        """
        events = self.get_filtered_events(**event_filters)
        subscriptions = Subscription.objects.filter(medium=self)

        subscription_q_objects = [
            Q(
                eventactor__entity__in=self.followed_by(sub.subscribed_entities()),
                source=sub.source
            )
            for sub in subscriptions if sub.only_following
        ]
        subscription_q_objects.append(
            Q(source__in=[sub.source for sub in subscriptions if not sub.only_following]))

        events = events.filter(reduce(or_, subscription_q_objects))
        return events

    @transaction.atomic
    def entity_events(self, entity, **event_filters):
        """Return subscribed events for a given entity.

        This method of getting events is useful when you want to see
        only the events relevant to a single entity. The events
        returned are events that the given entity is subscribed to,
        either directly as an individual entity, or because they are
        part of a group subscription. As an example, the
        `entity_events` method can be used to implement a newsfeed for
        a individual entity:

        .. code-block:: python

            TEMPLATE = '''
                <html><body>
                <h1> {entity}'s Events </h1>
                <ul>
                {% for event in events %}
                    <li> {{ event.context.event_text }} </li>
                {% endfor %}
                </ul>
                </body></html>
            '''

            def newsfeed(request):
                newsfeed_medium = Medium.objects.get(name='newsfeed')
                entity = Entity.get_for_obj(request.user)
                context = {}
                context['entity'] = entity
                context['events'] = site_feed_medium.entity_events(entity, seen=False, mark_seen=True)
                return HttpResponse(TEMPLATE.render(context))


        The only required argument for this method is the entity to
        get events for. Filtering based on the properties of the
        events themselves is supported, through the rest of the
        following arguments, which are optional.

        :type_entity: Entity
        :param entity: The entity to get events for.

        :type start_time: datetime.datetime (optional)
        :param start_time: Only return events that occurred after the
            given time. If no time is given for this argument, no
            filtering is done.

        :type end_time: datetime.datetime (optional)
        :param end_time: Only return events that occurred before the
            given time. If no time is given for this argument, no
            filtering is done

        :type seen: Boolean (optional)
        :param seen: This flag controls whether events that have
            marked as seen are included. By default, both events that
            have and have not been marked as seen are included. If
            ``True`` is given for this parameter, only events that
            have been marked as seen will be included. If ``False`` is
            given, only events that have not been marked as seen will
            be included.

        :type include_expired: Boolean (optional)
        :param include_expired: By default, events that have a
            expiration time, which has passed, are not included in the
            results. Passing in ``True`` to this argument causes
            expired events to be returned as well.

        :type actor: Entity (optional)
        :param actor: Only include events with the given entity as an
            actor.

        :type mark_seen: Boolean (optional)
        :param mark_seen: Create a side effect in the database that
            marks all the returned events as having been seen by this
            medium.

        :rtype: EventQuerySet
        :returns: A queryset of events.
        """
        events = self.get_filtered_events(**event_filters)

        subscriptions = Subscription.objects.filter(medium=self)
        subscriptions = self.subset_subscriptions(subscriptions, entity)

        subscription_q_objects = [
            Q(
                eventactor__entity__in=self.followed_by(entity),
                source=sub.source
            )
            for sub in subscriptions if sub.only_following
        ]
        subscription_q_objects.append(
            Q(source__in=[sub.source for sub in subscriptions if not sub.only_following])
        )

        return [
            event for event in events.filter(reduce(or_, subscription_q_objects))
            if self.filter_source_targets_by_unsubscription(event.source_id, [entity])
        ]

    @transaction.atomic
    def events_targets(self, entity_kind=None, **event_filters):
        """Return all events for this medium, with who each event is for.

        This method is useful for individually notifying every
        entity concerned with a collection of events, while
        still respecting subscriptions and usubscriptions. For
        example, ``events_targets`` can be used to send email
        notifications, by retrieving all unseen events (and marking
        them as now having been seen), and then processing the
        emails. In code, this could look like:

        .. code-block:: python

            email = Medium.objects.get(name='email')
            new_emails = email.events_targets(seen=False, mark_seen=True)

            for event, targets in new_emails:
                django.core.mail.send_mail(
                    subject = event.context["subject"]
                    message = event.context["message"]
                    recipient_list = [t.entity_meta["email"] for t in targets]
                )

        This ``events_targets`` method attempts to make bulk
        processing of push-style notifications straightforward. This
        sort of processing should normally occur in a separate thread
        from any request/response cycle.

        Filtering based on the properties of the events themselves is
        supported, through the rest of the following arguments, which
        are optional.

        :type entity_kind: EntityKind
        :param entity_kind: Only include targets of the given kind in
            each targets list.

        :type start_time: datetime.datetime (optional)
        :param start_time: Only return events that occurred after the
            given time. If no time is given for this argument, no
            filtering is done.

        :type end_time: datetime.datetime (optional)
        :param end_time: Only return events that occurred before the
            given time. If no time is given for this argument, no
            filtering is done

        :type seen: Boolean (optional)
        :param seen: This flag controls whether events that have
            marked as seen are included. By default, both events that
            have and have not been marked as seen are included. If
            ``True`` is given for this parameter, only events that
            have been marked as seen will be included. If ``False`` is
            given, only events that have not been marked as seen will
            be included.

        :type include_expired: Boolean (optional)
        :param include_expired: By default, events that have a
            expiration time, which has passed, are not included in the
            results. Passing in ``True`` to this argument causes
            expired events to be returned as well.

        :type actor: Entity (optional)
        :param actor: Only include events with the given entity as an
            actor.

        :type mark_seen: Boolean (optional)
        :param mark_seen: Create a side effect in the database that
            marks all the returned events as having been seen by this
            medium.

        :rtype: List of tuples
        :returns: A list of tuples in the form ``(event, targets)``
            where ``targets`` is a list of entities.
        """
        events = self.get_filtered_events(**event_filters)
        subscriptions = Subscription.objects.filter(medium=self)

        event_pairs = []
        for event in events:
            targets = []
            for sub in subscriptions:
                if event.source != sub.source:
                    continue

                subscribed = sub.subscribed_entities()
                if sub.only_following:
                    potential_targets = self.followers_of(
                        event.eventactor_set.values_list('entity__id', flat=True)
                    )
                    subscription_targets = list(Entity.objects.filter(
                        Q(id__in=subscribed), Q(id__in=potential_targets)))
                else:
                    subscription_targets = list(subscribed)

                targets.extend(subscription_targets)

            targets = self.filter_source_targets_by_unsubscription(event.source_id, targets)

            if entity_kind:
                targets = [t for t in targets if t.entity_kind == entity_kind]
            if targets:
                event_pairs.append((event, targets))

        return event_pairs

    def subset_subscriptions(self, subscriptions, entity=None):
        """Return only subscriptions the given entity is a part of.

        An entity is "part of a subscription" if either:

        1. The subscription is for that entity, with no
        sub-entity-kind. That is, it is not a group subscription.

        2. The subscription is for a super-entity of the given entity,
        and the subscription's sub-entity-kind is the same as that of
        the entity's.

        :type subscriptions: QuerySet
        :param subscriptions: A QuerySet of subscriptions to subset.

        :type entity: (optional) Entity
        :param entity: Subset subscriptions to only those relevant for
            this entity.

        :rtype: QuerySet
        :returns: A queryset of filtered subscriptions.
        """
        if entity is None:
            return subscriptions
        super_entities = EntityRelationship.objects.filter(
            sub_entity=entity).values_list('super_entity')
        subscriptions = subscriptions.filter(
            Q(entity=entity, sub_entity_kind=None) |
            Q(entity__in=super_entities, sub_entity_kind=entity.entity_kind)
        )

        return subscriptions

    @cached_property
    def unsubscriptions(self):
        """Returns the unsubscribed entity IDs for each source as a dict,
        keyed on source_id.

        :rtype: Dictionary
        :returns: A dictionary of the form ``{source_id: entities}``
            where ``entities`` is a list of entities unsubscribed from
            that source for this medium.
        """
        unsubscriptions = defaultdict(list)
        for unsub in Unsubscription.objects.filter(medium=self).values('entity', 'source'):
            unsubscriptions[unsub['source']].append(unsub['entity'])
        return unsubscriptions

    def filter_source_targets_by_unsubscription(self, source_id, targets):
        """Given a source id and targets, filter the targets by
        unsubscriptions. Return the filtered list of targets.
        """
        unsubscriptions = self.unsubscriptions
        return [t for t in targets if t.id not in unsubscriptions[source_id]]

    def get_filtered_events_queries(self, start_time, end_time, seen, include_expired, actor):
        """Return Q objects to filter events table to relevant events.

        The filters that are applied are those passed in from the
        method that is querying the events table: One of ``events``,
        ``entity_events`` or ``events_targets``. The arguments have
        the behavior documented in those methods.

        :rtype: List of Q objects
        :returns: A list of Q objects, which can be used as arguments
            to ``Event.objects.filter``.
        """
        now = datetime.utcnow()
        filters = []
        if start_time is not None:
            filters.append(Q(time__gte=start_time))
        if end_time is not None:
            filters.append(Q(time__lte=end_time))
        if not include_expired:
            filters.append(Q(time_expires__gte=now))

        # Check explicitly for True and False as opposed to None
        #   - `seen==False` gets unseen notifications
        #   - `seen is None` does no seen/unseen filtering
        if seen is True:
            filters.append(Q(eventseen__medium=self))
        elif seen is False:
            unseen_ids = _unseen_event_ids(medium=self)
            filters.append(Q(id__in=unseen_ids))

        # Filter by actor
        if actor is not None:
            filters.append(Q(eventactor__entity=actor))

        return filters

    def get_filtered_events(
            self, start_time=None, end_time=None, seen=None, mark_seen=False, include_expired=False, actor=None):
        """Retrieves events, filters by event level filters, and marks them as
        seen if necessary.

        :rtype: EventQuerySet
        :returns: All events which match the given filters.
        """
        filtered_events_queries = self.get_filtered_events_queries(start_time, end_time, seen, include_expired, actor)
        events = Event.objects.filter(*filtered_events_queries)
        if seen is False and mark_seen:
            # Evaluate the event qset here and create a new queryset that is no longer filtered by
            # if the events are marked as seen. We do this because we want to mark the events
            # as seen in the next line of code. If we didn't evaluate the qset here first, it result
            # in not returning unseen events since they are marked as seen.
            events = Event.objects.filter(id__in=list(e.id for e in events))
            events.mark_seen(self)

        return events

    def followed_by(self, entities):
        """Define what entities are followed by the entities passed to this
        method.

        This method can be overridden by a class that concretely
        inherits ``Medium``, to define custom semantics for the
        ``only_following`` flag on relevant ``Subscription``
        objects. Overriding this method, and ``followers_of`` will be
        sufficient to define that behavior. This method is not useful
        to call directly, but is used by the methods that filter
        events and targets.

        This implementation attempts to provide a sane default. In
        this implementation, the entities followed by the ``entities``
        argument are the entities themselves, and their super entities.

        That is, individual entities follow themselves, and the groups
        they are a part of. This works as a default implementation,
        but, for example, an alternate medium may wish to define the
        opposite behavior, where an individual entity follows
        themselves and all of their sub-entities.

        Return a queryset of the entities that the given entities are
        following. This needs to be the inverse of ``followers_of``.

        :type entities: Entity or EntityQuerySet
        :param entities: The Entity, or QuerySet of Entities of interest.

        :rtype: EntityQuerySet
        :returns: A QuerySet of all the entities followed by any of
            those given.
        """
        if isinstance(entities, Entity):
            entities = Entity.objects.filter(id=entities.id)
        super_entities = EntityRelationship.objects.filter(
            sub_entity__in=entities).values_list('super_entity')
        followed_by = Entity.objects.filter(
            Q(id__in=entities) | Q(id__in=super_entities))
        return followed_by

    def followers_of(self, entities):
        """Define what entities are followers of the entities passed to this
        method.

        This method can be overridden by a class that concretely
        inherits ``Medium``, to define custom semantics for the
        ``only_following`` flag on relevant ``Subscription``
        objects. Overriding this method, and ``followed_by`` will be
        sufficient to define that behavior. This method is not useful
        to call directly, but is used by the methods that filter
        events and targets.

        This implementation attempts to provide a sane default. In
        this implementation, the followers of the entities passed in
        are defined to be the entities themselves, and their
        sub-entities.

        That is, the followers of individual entities are themselves,
        and if the entity has sub-entities, those sub-entities. This
        works as a default implementation, but, for example, an
        alternate medium may wish to define the opposite behavior,
        where an the followers of an individual entity are themselves
        and all of their super-entities.

        Return a queryset of the entities that follow the given
        entities. This needs to be the inverse of ``followed_by``.

        :type entities: Entity or EntityQuerySet
        :param entities: The Entity, or QuerySet of Entities of interest.

        :rtype: EntityQuerySet
        :returns: A QuerySet of all the entities who are followers of
            any of those given.
        """
        if isinstance(entities, Entity):
            entities = Entity.objects.filter(id=entities.id)
        sub_entities = EntityRelationship.objects.filter(
            super_entity__in=entities).values_list('sub_entity')
        followers_of = Entity.objects.filter(
            Q(id__in=entities) | Q(id__in=sub_entities))
        return followers_of

    def render(self, events):
        """
        Renders a list of events for this medium. The events first have their contexts loaded.
        Afterwards, the rendered events are returned as a dictionary keyed on the event itself.
        The key points to a tuple of (txt, html) renderings of the event.

        :type events: list
        :param events: A list or queryset of Event models.

        :rtype: dict
        :returns: A dictionary of rendered text and html tuples keyed on the provided events.
        """
        from entity_event import context_loader
        context_loader.load_contexts_and_renderers(events, [self])
        return {e: e.render(self) for e in events}


@python_2_unicode_compatible
class Source(models.Model):
    """A ``Source`` is an object in the database that represents where
    events come from. These objects only require a few fields,
    ``name``, ``display_name`` ``description``, and ``group``.
    Source objects categorize events
    based on where they came from, or what type of information they
    contain. Each source should be fairly fine grained, with broader
    categorizations possible through ``SourceGroup`` objects. Sources
    can be created with ``Source.objects.create`` using the following
    parameters:

    :type name: str
    :param name: A short, unique name for the source.

    :type display_name: str
    :param display_name: A short, human readable name for the source.
        Does not need to be unique.

    :type description: str
    :param description: A human readable description of the source.

    :type group: SourceGroup
    :param group: A SourceGroup object. A broad grouping of where the
        events originate.

    Storing source objects in the database servers two purposes. The
    first is to provide an object that Subscriptions can reference,
    allowing different categories of events to be subscribed to over
    different mediums. The second is to allow source instances to
    store a reference to a function which can populate event contexts
    with additional information that is relevant to the source. This
    allows ``Event`` objects to be created with minimal data
    duplication.

    Once sources are created, they will primarily be used to
    categorize events, as each ``Event`` object requires a reference
    to a source. Additionally they will be referenced by
    ``Subscription`` objects to route events of the given source to be
    handled by a given medium.
    """
    name = models.CharField(max_length=64, unique=True)
    display_name = models.CharField(max_length=64)
    description = models.TextField()
    group = models.ForeignKey('SourceGroup')

    def __str__(self):
        """Readable representation of ``Source`` objects."""
        return self.display_name


@python_2_unicode_compatible
class SourceGroup(models.Model):
    """A ``SourceGroup`` object is a high level categorization of
    events. Since ``Source`` objects are meant to be very fine
    grained, they are collected into ``SourceGroup`` objects. There is
    no additional behavior associated with the source groups other
    than further categorization. Source groups can be created with
    ``SourceGroup.objects.create``, which takes the following
    arguments:

    :type name: str
    :param name: A short, unique name for the source group.

    :type display_name: str
    :param display_name: A short, human readable name for the source
        group. Does not need to be unique.

    :type description: str
    :param description: A human readable description of the source
        group.
    """
    name = models.CharField(max_length=64, unique=True)
    display_name = models.CharField(max_length=64)
    description = models.TextField()

    def __str__(self):
        """Readable representation of ``SourceGroup`` objects."""
        return self.display_name


@python_2_unicode_compatible
class Unsubscription(models.Model):
    """Because django-entity-event allows for whole groups to be
    subscribed to events at once, unsubscribing an entity is not as
    simple as removing their subscription object. Instead, the
    Unsubscription table provides a simple way to ensure that an
    entity does not see events if they don't want to.

    Unsubscriptions are created for a single entity at a time, where
    they are unsubscribed for events from a source on a medium. This
    is stored as an ``Unsubscription`` object in the database, which
    can be created using ``Unsubscription.objects.create`` using the
    following arguments:

    :type entity: Entity
    :param entity: The entity to unsubscribe.

    :type medium: Medium
    :param medium: The ``Medium`` object representing where they don't
        want to see the events.

    :type source: Source
    :param source: The ``Source`` object representing what category
        of event they no longer want to see.

    Once an ``Unsubscription`` object is created, all of the logic to
    ensure that they do not see events form the given source by the
    given medium is handled by the methods used to query for events
    via the ``Medium`` object. That is, once the object is created, no
    more work is needed to unsubscribe them.
    """
    entity = models.ForeignKey(Entity)
    medium = models.ForeignKey('Medium')
    source = models.ForeignKey('Source')

    def __str__(self):
        """Readable representation of ``Unsubscription`` objects."""
        s = '{entity} from {source} by {medium}'
        entity = self.entity.__str__()
        source = self.source.__str__()
        medium = self.medium.__str__()
        return s.format(entity=entity, source=source, medium=medium)


@python_2_unicode_compatible
class Subscription(models.Model):
    """Which types of events are available to which mediums is controlled
    through ``Subscription`` objects. By creating a ``Subscription``
    object in the database, you are storing that events from a given
    ``Source`` object should be available to a given ``Medium``
    object.

    Each ``Subscription`` object can be one of two levels, either an
    individual subscription or a group subscription. Additionally,
    each ``Subscription`` object can be one of two types of
    subscription, either a global subscription, or an "only following"
    subscription. ``Subscription`` objects are created using
    ``Subscription.objects.create`` which takes the following
    arguments:

    :type medium: Medium
    :param medium: The ``Medium`` object to make events available to.

    :type source: Source
    :param source: The ``Source`` object that represents the category
        of events to make available.

    :type entity: Entity
    :param entity: The entity to subscribe in the case of an
        individual subscription, or in the case of a group
        subscription, the super-entity of the group.

    :type sub_entity_kind: (optional) EntityKind
    :param sub_entity_kind: When creating a group subscription, this
        is a foreign key to the ``EntityKind`` of the sub-entities to
        subscribe. In the case of an individual subscription, this should
        be ``None``.

    :type only_following: Boolean
    :param only_following: If ``True``, events will be available to
        entities through the medium only if the entities are following
        the actors of the event. If ``False``, the events will all
        be available to all the entities through the medium.

    When a ``Medium`` object is used to query for events, only the
    events that have a subscription for their source to that medium
    will ever be returned. This is an extremely useful property that
    allows complex subscription logic to be handled simply by storing
    subscription objects in the database.

    Storing subscriptions is made simpler by the ability to subscribe
    groups of entities with a single subscription object.  Groups of
    entities of a given kind can be subscribed by subscribing their
    super-entity and providing the ``sub_entity_kind`` argument.

    Subscriptions further are specified to be either an "only following"
    subscription or not. This specification controls what
    events will be returned when ``Medium.entity_events`` is called,
    and controls what targets are returned when
    ``Medium.events_targets`` is called.

    For example, if events are created for a new photo being uploaded
    (from a single source called, say "photos"), and we want to provide
    individuals with a notification in their newsfeed (through a
    medium called "newsfeed"), we want to be able to display only the
    events where the individual is tagged in the photo. By setting
    ``only_following`` to true the following code would only return
    events where the individual was included in the ``EventActor`` s,
    rather than returning all "photos" events:

    .. code-block:: python

        user_entity = Entity.objects.get_for_obj(user)
        newsfeed_medium = Medium.objects.get(name='newsfeed')
        newsfeed.entity_events(user)

    The behavior of what constitutes "following" is controlled by the
    Medium class. A default implementation of following is provided
    and documented in the ``Medium.followers_of`` and
    ``Medium.followed_by`` methods, but could be extended by
    subclasses of Medium.
    """
    medium = models.ForeignKey('Medium')
    source = models.ForeignKey('Source')
    entity = models.ForeignKey(Entity, related_name='+')
    sub_entity_kind = models.ForeignKey(EntityKind, null=True, related_name='+', default=None)
    only_following = models.BooleanField(default=True)

    def __str__(self):
        """Readable representation of ``Subscription`` objects."""
        s = '{entity} to {source} by {medium}'
        entity = self.entity.__str__()
        source = self.source.__str__()
        medium = self.medium.__str__()
        return s.format(entity=entity, source=source, medium=medium)

    def subscribed_entities(self):
        """Return a queryset of all subscribed entities.

        This will be a single entity in the case of an individual
        subscription, otherwise it will be all the entities in the
        group subscription.

        :rtype: EntityQuerySet
        :returns: A QuerySet of all the entities that are a part of
            this subscription.
        """
        if self.sub_entity_kind is not None:
            sub_entities = self.entity.sub_relationships.filter(
                sub_entity__entity_kind=self.sub_entity_kind).values_list('sub_entity')
            entities = Entity.objects.filter(id__in=sub_entities)
        else:
            entities = Entity.objects.filter(id=self.entity.id)
        return entities


class EventQuerySet(QuerySet):
    """A custom QuerySet for Events.
    """
    def mark_seen(self, medium):
        """Creates EventSeen objects for the provided medium for every event
        in the queryset.

        Creating these EventSeen objects ensures they will not be
        returned when passing ``seen=False`` to any of the medium
        event retrieval functions, ``events``, ``entity_events``, or
        ``events_targets``.
        """
        EventSeen.objects.bulk_create([
            EventSeen(event=event, medium=medium) for event in self
        ])

    def load_contexts_and_renderers(self, medium):
        """
        Loads context data into the event ``context`` variable. This method
        destroys the queryset and returns a list of events.
        """
        from entity_event import context_loader
        return context_loader.load_contexts_and_renderers(self, [medium])


class EventManager(models.Manager):
    """A custom Manager for Events.
    """
    def get_queryset(self):
        """Return the EventQuerySet.
        """
        return EventQuerySet(self.model)

    def mark_seen(self, medium):
        """Creates EventSeen objects for the provided medium for every event
        in the queryset.

        Creating these EventSeen objects ensures they will not be
        returned when passing ``seen=False`` to any of the medium
        event retrieval functions, ``events``, ``entity_events``, or
        ``events_targets``.
        """
        return self.get_queryset().mark_seen(medium)

    def load_contexts_and_renderers(self, medium):
        """
        Loads context data into the event ``context`` variable. This method
        destroys the queryset and returns a list of events.
        """
        return self.get_queryset().load_contexts_and_renderers(medium)

    @transaction.atomic
    def create_event(self, actors=None, ignore_duplicates=False, **kwargs):
        """Create events with actors.

        This method can be used in place of ``Event.objects.create``
        to create events, and the appropriate actors. It takes all the
        same keywords as ``Event.objects.create`` for the event
        creation, but additionally takes a list of actors, and can be
        told to not attempt to create an event if a duplicate event
        exists.

        :type source: Source
        :param source: A ``Source`` object representing where the
            event came from.

        :type context: dict
        :param context: A dictionary containing relevant
            information about the event, to be serialized into
            JSON. It is possible to load additional context
            dynamically  when events are fetched. See the
            documentation on the ``ContextRenderer`` model.

        :type uuid: str
        :param uuid: A unique string for the event. Requiring a
            ``uuid`` allows code that creates events to ensure they do
            not create duplicate events. This id could be, for example
            some hash of the ``context``, or, if the creator is
            unconcerned with creating duplicate events a call to
            python's ``uuid1()`` in the ``uuid`` module.

        :type time_expires: datetime (optional)
        :param time_expires: If given, the default methods for
            querying events will not return this event after this time
            has passed.

        :type actors: (optional) List of entities or list of entity ids.
        :param actors: An ``EventActor`` object will be created for
            each entity in the list. This allows for subscriptions
            which are only following certain entities to behave
            appropriately.

        :type ignore_duplicates: (optional) Boolean
        :param ignore_duplicates: If ``True``, a check will be made to
            ensure that an event with the give ``uuid`` does not exist
            before attempting to create the event. Setting this to
            ``True`` allows the creator of events to gracefully ensure
            no duplicates are created.

        :rtype: Event
        :returns: The created event. Alternatively if a duplicate
            event already exists and ``ignore_duplicates`` is
            ``True``, it will return ``None``.
        """
        if ignore_duplicates and self.filter(uuid=kwargs.get('uuid', '')).exists():
            return None

        event = self.create(**kwargs)

        # Allow user to pass pks for actors
        actors = [
            a.id if isinstance(a, Entity) else a
            for a in actors
        ] if actors else []

        EventActor.objects.bulk_create([EventActor(entity_id=actor, event=event) for actor in actors])
        return event


@python_2_unicode_compatible
class Event(models.Model):
    """``Event`` objects store information about events. By storing
    events, from a given source, with some context, they are made
    available to any ``Medium`` object with an appropriate
    subscription. Events can be created with
    ``Event.objects.create_event``, documented above.

    When creating an event, the information about what occurred is
    stored in a JSON blob in the ``context`` field. This context can
    be any type of information that could be useful for displaying
    events on a given Medium. It is entirely the role of the
    application developer to ensure that there is agreement between
    what information is stored in ``Event.context`` and what
    information the code the processes and displays events on each
    medium expects.

    Events will usually be created by code that also created, or knows
    about the ``Source`` object that is required to create the event.
    To prevent storing unnecessary data in the context, this code can
    define a context loader function when creating this source, which
    can be used to dynamically fetch more data based on whatever
    limited amount of data makes sense to store in the context. This
    is further documented in the ``Source`` documentation.
    """
    source = models.ForeignKey('Source')
    context = jsonfield.JSONField()
    time = models.DateTimeField(auto_now_add=True, db_index=True)
    time_expires = models.DateTimeField(default=datetime.max, db_index=True)
    uuid = models.CharField(max_length=128, unique=True)

    objects = EventManager()

    def __init__(self, *args, **kwargs):
        super(Event, self).__init__(*args, **kwargs)
        # A dictionary that is populated with renderers after the contexts have been
        # properly loaded. When renderers are available, the 'render' method may be
        # called with a medium and optional observer
        self._context_renderers = {}

    def render(self, medium, observing_entity=None):
        """
        Returns the rendered event as a tuple of text and html content. This information
        is filled out with respect to which medium is rendering the event, what context
        renderers are available with the prefetched context, and which optional entity
        may be observing the rendered event.
        """
        if medium not in self._context_renderers:
            raise RuntimeError('Context and renderer for medium {0} has not or cannot been fetched'.format(medium))
        else:
            return self._context_renderers[medium].render_context_to_text_html_templates(self.context)

    def __str__(self):
        """Readable representation of ``Event`` objects."""
        s = '{source} event at {time}'
        source = self.source.__str__()
        time = self.time.strftime('%Y-%m-%d::%H:%M:%S')
        return s.format(source=source, time=time)


class AdminEvent(Event):
    """A proxy model used to provide a separate interface for event
    creation through the django-admin interface.
    """
    class Meta:
        proxy = True


@python_2_unicode_compatible
class EventActor(models.Model):
    """``EventActor`` objects encode what entities were involved in an
    event. They provide the information necessary to create "only
    following" subscriptions which route events only to the entities
    that are involved in the event.

    ``EventActor`` objects should not be created directly, but should
    be created as part of the creation of ``Event`` objects, using
    ``Event.objects.create_event``.
    """
    event = models.ForeignKey('Event')
    entity = models.ForeignKey(Entity)

    def __str__(self):
        """Readable representation of ``EventActor`` objects."""
        s = 'Event {eventid} - {entity}'
        eventid = self.event.id
        entity = self.entity.__str__()
        return s.format(eventid=eventid, entity=entity)


@python_2_unicode_compatible
class EventSeen(models.Model):
    """``EventSeen`` objects store information about where and when an
    event was seen. They store the medium that the event was seen on,
    and what time it was seen. This information is used by the event
    querying methods on ``Medium`` objects to filter events by whether
    or not they have been seen on that medium.

    ``EventSeen`` objects should not be created directly, but should
    be created by using the ``EventQuerySet.mark_seen`` method,
    available on the QuerySets returned by the event querying methods.
    """
    event = models.ForeignKey('Event')
    medium = models.ForeignKey('Medium')
    time_seen = models.DateTimeField(default=datetime.utcnow)

    class Meta:
        unique_together = ('event', 'medium')

    def __str__(self):
        """Readable representation of ``EventSeen`` objects."""
        s = 'Seen on {medium} at {time}'
        medium = self.medium.__str__()
        time = self.time_seen.strftime('%Y-%m-%d::%H:%M:%S')
        return s.format(medium=medium, time=time)


def _unseen_event_ids(medium):
    """Return all events that have not been seen on this medium.
    """
    query = '''
    SELECT event.id
    FROM entity_event_event AS event
        LEFT OUTER JOIN (SELECT *
                         FROM entity_event_eventseen AS seen
                         WHERE seen.medium_id=%s) AS eventseen
            ON event.id = eventseen.event_id
    WHERE eventseen.medium_id IS NULL
    '''
    unseen_events = Event.objects.raw(query, params=[medium.id])
    ids = [e.id for e in unseen_events]
    return ids


@python_2_unicode_compatible
class RenderingStyle(models.Model):
    """
    Defines a rendering style. This is used to group together mediums that have
    similar rendering styles and allows context renderers to be used across
    mediums.
    """
    name = models.CharField(max_length=64, unique=True)
    display_name = models.CharField(max_length=64, default='')

    def __str__(self):
        return self.display_name


@python_2_unicode_compatible
class ContextRenderer(models.Model):
    """``ContextRenderer`` objects store information about how
    a source or source group is rendered with a particular rendering style, along with
    information for loading the render context in a database-efficient
    manner.

    Of the four template fields: `text_template_path`, 'html_template_path',
    `text_template`, and `html_template`, at least one must be
    non-empty. Both a text and html template may be provided, either
    through a path to the template, or a raw template object.
    If both are provided, the template given in the path will be used and
    the text template will be ignored.

    This object is linked to a `RenderingStyle` object. This is how the
    context renderer is associated with various `Medium` objects. It also
    provides the `source` that uses the renderer. If a `source_group` is specified,
    all sources under that group use this context renderer for the rendering style.

    The `context_hints` provide the ability to fetch model IDs of an event context that
    are stored in the database. For example, if an event context has a `user` key that
    points to the PK of a Django `User` model, the context hints for it would be specified
    as follows:

    .. code-block:: python

        {
            'user': {
                'app_name': 'auth',
                'model_name': 'User',
            }
        }

    With these hints, the 'user' field in the event context will be treated as a PK in the
    database and fetched appropriately. If one wishes to perform and prefetch or select_related
    calls, the following options can be added:

    .. code-block:: python

        {
            'user': {
                'app_name': 'auth',
                'model_name': 'User',
                'select_related': ['foreign_key_field', 'one_to_one_field'],
                'prefetch_related': ['reverse_foreign_key_field', 'many_to_many_field'],
            }
        }

    Note that as many keys can be defined that have corresponding keys in the event context for
    the particular source or source group. Also note that the keys in the event context can
    be embedded anywhere in the context and can also point to a list of PKs. For example:

    .. code-block:: python

        {
            'my_context': {
                'user': [1, 3, 5, 10],
                'other_context_info': 'other_info_string',
            },
            'more_context': {
                'hello': 'world',
            }
        }

    In the above case, `User` objects with the PKs 1, 3, 5, and 10 will be fetched and loaded into
    the event context whenever rendering is performed.
    """
    name = models.CharField(max_length=64, unique=True)

    # The various templates that can be used for rendering
    text_template_path = models.CharField(max_length=256, default='')
    html_template_path = models.CharField(max_length=256, default='')
    text_template = models.TextField(default='')
    html_template = models.TextField(default='')

    # The source or source group of the event. It can only be one or the other
    source = models.ForeignKey(Source, null=True)
    source_group = models.ForeignKey(SourceGroup, null=True)

    # The rendering style. Used to associated it with a medium
    rendering_style = models.ForeignKey(RenderingStyle)

    # Containts hints on how to fetch the context from the database
    context_hints = jsonfield.JSONField(null=True, default=None)

    class Meta:
        unique_together = ('source', 'rendering_style')

    def get_sources(self):
        return [self.source] if self.source_id else self.source_group.source_set.all()

    def __str__(self):
        return self.name

    def render_text_or_html_template(self, context, is_text=True):
        """
        Renders a text or html template based on either the template path or the
        stored template.
        """
        template_path = getattr(self, '{0}_template_path'.format('text' if is_text else 'html'))
        template = getattr(self, '{0}_template'.format('text' if is_text else 'html'))
        if template_path:
            return render_to_string(template_path, context)
        elif template:
            return Template(template).render(Context(context))
        else:
            return ''

    def render_context_to_text_html_templates(self, context):
        """Render the templates with the provided context.

        Args:
          A loaded context.

        Returns:
          A tuple of (rendered_text, rendered_html). Either, but not both
          may be an empty string.
        """
        # Process text template:
        return (
            self.render_text_or_html_template(context, is_text=True).strip(),
            self.render_text_or_html_template(context, is_text=False).strip(),
        )



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding index on 'Event', fields ['time']
        db.create_index(u'entity_event_event', ['time'])


        # Changing field 'Event.time_expires'
        db.alter_column(u'entity_event_event', 'time_expires', self.gf('django.db.models.fields.DateTimeField')())
        # Adding index on 'Event', fields ['time_expires']
        db.create_index(u'entity_event_event', ['time_expires'])


    def backwards(self, orm):
        # Removing index on 'Event', fields ['time_expires']
        db.delete_index(u'entity_event_event', ['time_expires'])

        # Removing index on 'Event', fields ['time']
        db.delete_index(u'entity_event_event', ['time'])


        # Changing field 'Event.time_expires'
        db.alter_column(u'entity_event_event', 'time_expires', self.gf('django.db.models.fields.DateTimeField')(null=True))

    models = {
        u'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        u'entity.entity': {
            'Meta': {'unique_together': "(('entity_id', 'entity_type', 'entity_kind'),)", 'object_name': 'Entity'},
            'display_name': ('django.db.models.fields.TextField', [], {'db_index': 'True', 'blank': 'True'}),
            'entity_id': ('django.db.models.fields.IntegerField', [], {}),
            'entity_kind': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity.EntityKind']", 'on_delete': 'models.PROTECT'}),
            'entity_meta': ('jsonfield.fields.JSONField', [], {'null': 'True'}),
            'entity_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['contenttypes.ContentType']", 'on_delete': 'models.PROTECT'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True', 'db_index': 'True'})
        },
        u'entity.entitykind': {
            'Meta': {'object_name': 'EntityKind'},
            'display_name': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '256', 'db_index': 'True'})
        },
        u'entity_event.event': {
            'Meta': {'object_name': 'Event'},
            'context': ('jsonfield.fields.JSONField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'source': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity_event.Source']"}),
            'time': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'db_index': 'True', 'blank': 'True'}),
            'time_expires': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime(9999, 12, 31, 0, 0)', 'db_index': 'True'}),
            'uuid': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '128'})
        },
        u'entity_event.eventactor': {
            'Meta': {'object_name': 'EventActor'},
            'entity': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity.Entity']"}),
            'event': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity_event.Event']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'})
        },
        u'entity_event.eventseen': {
            'Meta': {'unique_together': "(('event', 'medium'),)", 'object_name': 'EventSeen'},
            'event': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity_event.Event']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'medium': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity_event.Medium']"}),
            'time_seen': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.utcnow'})
        },
        u'entity_event.medium': {
            'Meta': {'object_name': 'Medium'},
            'description': ('django.db.models.fields.TextField', [], {}),
            'display_name': ('django.db.models.fields.CharField', [], {'max_length': '64'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '64'})
        },
        u'entity_event.source': {
            'Meta': {'object_name': 'Source'},
            'context_loader': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '256', 'blank': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {}),
            'display_name': ('django.db.models.fields.CharField', [], {'max_length': '64'}),
            'group': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity_event.SourceGroup']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '64'})
        },
        u'entity_event.sourcegroup': {
            'Meta': {'object_name': 'SourceGroup'},
            'description': ('django.db.models.fields.TextField', [], {}),
            'display_name': ('django.db.models.fields.CharField', [], {'max_length': '64'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '64'})
        },
        u'entity_event.subscription': {
            'Meta': {'object_name': 'Subscription'},
            'entity': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': u"orm['entity.Entity']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'medium': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity_event.Medium']"}),
            'only_following': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'source': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity_event.Source']"}),
            'sub_entity_kind': ('django.db.models.fields.related.ForeignKey', [], {'default': 'None', 'related_name': "'+'", 'null': 'True', 'to': u"orm['entity.EntityKind']"})
        },
        u'entity_event.unsubscription': {
            'Meta': {'object_name': 'Unsubscription'},
            'entity': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity.Entity']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'medium': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity_event.Medium']"}),
            'source': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['entity_event.Source']"})
        }
    }

    complete_apps = ['entity_event']


from django.db import models


class TestFKModel(models.Model):
    value = models.CharField(max_length=64)


class TestFKModel2(models.Model):
    value = models.CharField(max_length=64)


class TestModel(models.Model):
    value = models.CharField(max_length=64)
    fk = models.ForeignKey(TestFKModel)
    fk2 = models.ForeignKey(TestFKModel2)
    fk_m2m = models.ManyToManyField(TestFKModel, related_name='+')



from django.test import SimpleTestCase, TestCase
from django_dynamic_fixture import N, G
from mock import patch

from entity_event import context_loader
from entity_event import models
from entity_event.tests import models as test_models


class TestGetContextHintsFromSource(SimpleTestCase):
    def test_no_context_renderers(self):
        res = context_loader.get_context_hints_per_source([])
        self.assertEquals(res, {})

    @patch.object(models.ContextRenderer, 'get_sources', spec_set=True)
    def test_one_context_renderer(self, mock_get_sources):
        source = N(models.Source, persist_dependencies=False)
        mock_get_sources.return_value = [source]
        res = context_loader.get_context_hints_per_source([
            N(models.ContextRenderer, source=source, context_hints={
                'key': {
                    'app_name': 'entity_event.tests',
                    'model_name': 'TestModel',
                    'select_related': ['fk'],
                },
            }, persist_dependencies=False)
        ])
        self.assertEquals(res, {
            source: {
                'key': {
                    'app_name': 'entity_event.tests',
                    'model_name': 'TestModel',
                    'select_related': set(['fk']),
                    'prefetch_related': set(),
                }
            }
        })

    def test_multiple_context_renderers_over_multiple_source(self):
        source1 = N(models.Source, persist_dependencies=False, id=1)
        source2 = N(models.Source, persist_dependencies=False, id=2)
        res = context_loader.get_context_hints_per_source([
            N(models.ContextRenderer, source=source1, context_hints={
                'key': {
                    'app_name': 'entity_event.tests',
                    'model_name': 'TestModel',
                    'select_related': ['fk'],
                    'prefetch_related': ['prefetch1', 'prefetch2'],
                },
            }, persist_dependencies=False),
            N(models.ContextRenderer, source=source1, context_hints={
                'key': {
                    'app_name': 'entity_event.tests',
                    'model_name': 'TestModel',
                    'select_related': ['fk1'],
                    'prefetch_related': ['prefetch2', 'prefetch3'],
                },
            }, persist_dependencies=False),
            N(models.ContextRenderer, source=source2, context_hints={
                'key2': {
                    'app_name': 'entity_event.tests2',
                    'model_name': 'TestModel2',
                    'select_related': ['fk2'],
                    'prefetch_related': ['prefetch5', 'prefetch6'],
                },
            }, persist_dependencies=False)
        ])
        self.assertEquals(res, {
            source1: {
                'key': {
                    'app_name': 'entity_event.tests',
                    'model_name': 'TestModel',
                    'select_related': set(['fk', 'fk1']),
                    'prefetch_related': set(['prefetch1', 'prefetch2', 'prefetch3']),
                }
            },
            source2: {
                'key2': {
                    'app_name': 'entity_event.tests2',
                    'model_name': 'TestModel2',
                    'select_related': set(['fk2']),
                    'prefetch_related': set(['prefetch5', 'prefetch6']),
                }
            },
        })


class TestGetQuerysetsForContextHints(SimpleTestCase):
    def test_no_context_hints(self):
        qsets = context_loader.get_querysets_for_context_hints({})
        self.assertEquals(qsets, {})

    def test_one_context_hint_no_select_related(self):
        source = N(models.Source, persist_dependencies=False, id=1)
        qsets = context_loader.get_querysets_for_context_hints({
            source: {
                'key': {
                    'app_name': 'tests',
                    'model_name': 'TestModel',
                },
            },
        })
        self.assertEquals(qsets, {
            test_models.TestModel: test_models.TestModel.objects
        })

    def test_one_context_hint_w_select_related(self):
        source = N(models.Source, persist_dependencies=False, id=1)
        qsets = context_loader.get_querysets_for_context_hints({
            source: {
                'key': {
                    'app_name': 'tests',
                    'model_name': 'TestModel',
                    'select_related': ['fk'],
                },
            },
        })
        # Verify the raw sql to ensure select relateds will happen
        raw_sql = (
            'SELECT "tests_testmodel"."id", "tests_testmodel"."value", "tests_testmodel"."fk_id", '
            '"tests_testmodel"."fk2_id", '
            '"tests_testfkmodel"."id", "tests_testfkmodel"."value" FROM "tests_testmodel" INNER JOIN '
            '"tests_testfkmodel" ON ( "tests_testmodel"."fk_id" = "tests_testfkmodel"."id" )'
        )

        self.assertEquals(str(qsets[test_models.TestModel].query), raw_sql)

    def test_multiple_context_hints_w_multiple_select_related(self):
        source = N(models.Source, persist_dependencies=False, id=1)
        source2 = N(models.Source, persist_dependencies=False, id=2)
        qsets = context_loader.get_querysets_for_context_hints({
            source: {
                'key': {
                    'app_name': 'tests',
                    'model_name': 'TestModel',
                    'select_related': ['fk'],
                },
            },
            source2: {
                'key2': {
                    'app_name': 'tests',
                    'model_name': 'TestModel',
                    'select_related': ['fk2'],
                },
            }
        })
        # Verify the raw sql to ensure select relateds will happen
        raw_sql = (
            'SELECT "tests_testmodel"."id", "tests_testmodel"."value", "tests_testmodel"."fk_id", '
            '"tests_testmodel"."fk2_id", '
            '"tests_testfkmodel"."id", "tests_testfkmodel"."value", '
            '"tests_testfkmodel2"."id", "tests_testfkmodel2"."value" FROM "tests_testmodel" INNER JOIN '
            '"tests_testfkmodel" ON ( "tests_testmodel"."fk_id" = "tests_testfkmodel"."id" ) '
            'INNER JOIN "tests_testfkmodel2" ON ( "tests_testmodel"."fk2_id" = "tests_testfkmodel2"."id" )'
        )

        self.assertEquals(str(qsets[test_models.TestModel].query), raw_sql)

    def test_multiple_context_hints_w_multiple_select_related_multiple_prefetch_related(self):
        source = N(models.Source, persist_dependencies=False, id=1)
        source2 = N(models.Source, persist_dependencies=False, id=2)
        qsets = context_loader.get_querysets_for_context_hints({
            source: {
                'key': {
                    'app_name': 'tests',
                    'model_name': 'TestModel',
                    'select_related': ['fk'],
                    'prefetch_related': ['fk_m2m'],
                },
            },
            source2: {
                'key2': {
                    'app_name': 'tests',
                    'model_name': 'TestModel',
                    'select_related': ['fk2'],
                },
            }
        })

        # Verify the raw sql to ensure select relateds will happen. Note that prefetch relateds are not
        # included in raw sql
        raw_sql = (
            'SELECT "tests_testmodel"."id", "tests_testmodel"."value", "tests_testmodel"."fk_id", '
            '"tests_testmodel"."fk2_id", '
            '"tests_testfkmodel"."id", "tests_testfkmodel"."value", '
            '"tests_testfkmodel2"."id", "tests_testfkmodel2"."value" FROM "tests_testmodel" INNER JOIN '
            '"tests_testfkmodel" ON ( "tests_testmodel"."fk_id" = "tests_testfkmodel"."id" ) '
            'INNER JOIN "tests_testfkmodel2" ON ( "tests_testmodel"."fk2_id" = "tests_testfkmodel2"."id" )'
        )

        self.assertEquals(str(qsets[test_models.TestModel].query), raw_sql)


class TestGetQuerysetsForContextHintsDbTests(TestCase):
    def test_multiple_context_hints_w_multiple_select_related_multiple_prefetch_related(self):
        source = N(models.Source, persist_dependencies=False, id=1)
        source2 = N(models.Source, persist_dependencies=False, id=2)
        qsets = context_loader.get_querysets_for_context_hints({
            source: {
                'key': {
                    'app_name': 'tests',
                    'model_name': 'TestModel',
                    'select_related': ['fk'],
                    'prefetch_related': ['fk_m2m'],
                },
            },
            source2: {
                'key2': {
                    'app_name': 'tests',
                    'model_name': 'TestModel',
                    'select_related': ['fk2'],
                },
            }
        })

        # Create objects to query in order to test optimal number of queries
        fk = G(test_models.TestFKModel)
        fk2 = G(test_models.TestFKModel2)
        o = G(test_models.TestModel, fk=fk, fk2=fk2)
        m2ms = [G(test_models.TestFKModel), G(test_models.TestFKModel)]
        o.fk_m2m.add(*m2ms)

        with self.assertNumQueries(2):
            v = qsets[test_models.TestModel].get(id=o.id)
            self.assertEquals(v.fk, fk)
            self.assertEquals(v.fk2, fk2)
            self.assertEquals(set(v.fk_m2m.all()), set(m2ms))


class DictFindTest(SimpleTestCase):
    def test_dict_find_none(self):
        self.assertEquals(list(context_loader.dict_find({}, 'key')), [])

    def test_dict_find_list_key(self):
        d = {'key': ['value']}
        self.assertEquals(list(context_loader.dict_find(d, 'key')), [(d, ['value'])])

    def test_dict_find_nested_list_key(self):
        d = {'key': ['value']}
        larger_dict = {
            'l': [{
                'hi': {},
            }, {
                'hi2': d
            }]
        }
        self.assertEquals(list(context_loader.dict_find(larger_dict, 'key')), [(d, ['value'])])

    def test_dict_find_double_nested_list_key(self):
        d = {'key': ['value']}
        larger_dict = {
            'l': [{
                'hi': {},
            }, {
                'hi2': d
            }],
            'hi3': d
        }
        self.assertEquals(list(context_loader.dict_find(larger_dict, 'key')), [(d, ['value']), (d, ['value'])])

    def test_dict_find_deep_nested_list_key(self):
        d = {'key': ['value']}
        larger_dict = [[{
            'l': [{
                'hi': {},
            }, {
                'hi2': d
            }],
            'hi3': d
        }]]
        self.assertEquals(list(context_loader.dict_find(larger_dict, 'key')), [(d, ['value']), (d, ['value'])])


class GetModelIdsToFetchTest(SimpleTestCase):
    def test_no_events(self):
        self.assertEquals(context_loader.get_model_ids_to_fetch([], {}), {})

    def test_no_context_hints(self):
        e = N(models.Event, context={}, persist_dependencies=False)
        self.assertEquals(context_loader.get_model_ids_to_fetch([e], {}), {})

    def test_w_one_event_one_context_hint_single_pk(self):
        source = N(models.Source, persist_dependencies=False, id=1)
        hints = {
            source: {
                'key': {
                    'app_name': 'tests',
                    'model_name': 'TestModel',
                },
            },
        }
        e = N(models.Event, context={'key': 2}, persist_dependencies=False, source=source)
        self.assertEquals(context_loader.get_model_ids_to_fetch([e], hints), {
            test_models.TestModel: set([2])
        })

    def test_w_one_event_one_context_hint_list_pks(self):
        source = N(models.Source, persist_dependencies=False, id=1)
        hints = {
            source: {
                'key': {
                    'app_name': 'tests',
                    'model_name': 'TestModel',
                },
            },
        }
        e = N(models.Event, context={'key': [2, 3, 5]}, persist_dependencies=False, source=source)
        self.assertEquals(context_loader.get_model_ids_to_fetch([e], hints), {
            test_models.TestModel: set([2, 3, 5])
        })

    def test_w_multiple_events_one_context_hint_list_pks(self):
        source = N(models.Source, persist_dependencies=False, id=1)
        hints = {
            source: {
                'key': {
                    'app_name': 'tests',
                    'model_name': 'TestModel',
                },
            },
        }
        e1 = N(models.Event, context={'key': [2, 3, 5]}, persist_dependencies=False, source=source)
        e2 = N(models.Event, context={'key': 88}, persist_dependencies=False, source=source)
        self.assertEquals(context_loader.get_model_ids_to_fetch([e1, e2], hints), {
            test_models.TestModel: set([2, 3, 5, 88])
        })

    def test_w_multiple_events_multiple_context_hints_list_pks(self):
        source1 = N(models.Source, persist_dependencies=False, id=1)
        source2 = N(models.Source, persist_dependencies=False, id=2)
        hints = {
            source1: {
                'key': {
                    'app_name': 'tests',
                    'model_name': 'TestModel',
                },
            },
            source2: {
                'key': {
                    'app_name': 'tests',
                    'model_name': 'TestModel',
                },
                'key2': {
                    'app_name': 'tests',
                    'model_name': 'TestFKModel',
                },
            },
        }
        e1 = N(models.Event, context={'key': [2, 3, 5]}, persist_dependencies=False, source=source1)
        e2 = N(models.Event, context={'key': 88}, persist_dependencies=False, source=source1)
        e3 = N(models.Event, context={'key': 100, 'key2': [50]}, persist_dependencies=False, source=source2)
        e4 = N(models.Event, context={'key2': [60]}, persist_dependencies=False, source=source2)
        self.assertEquals(context_loader.get_model_ids_to_fetch([e1, e2, e3, e4], hints), {
            test_models.TestModel: set([2, 3, 5, 88, 100]),
            test_models.TestFKModel: set([50, 60])
        })


class FetchModelDataTest(TestCase):
    def test_none(self):
        self.assertEquals({}, context_loader.fetch_model_data({}, {}))

    def test_one_model_one_id_to_fetch(self):
        m1 = G(test_models.TestModel)
        self.assertEquals({
            test_models.TestModel: {m1.id: m1}
        }, context_loader.fetch_model_data({
            test_models.TestModel: test_models.TestModel.objects
        }, {
            test_models.TestModel: [m1.id]
        }))

    def test_multiple_models_multiple_ids_to_fetch(self):
        m1 = G(test_models.TestModel)
        m2 = G(test_models.TestModel)
        m3 = G(test_models.TestFKModel)
        self.assertEquals({
            test_models.TestModel: {m1.id: m1, m2.id: m2},
            test_models.TestFKModel: {m3.id: m3}
        }, context_loader.fetch_model_data({
            test_models.TestModel: test_models.TestModel.objects,
            test_models.TestFKModel: test_models.TestFKModel.objects,
        }, {
            test_models.TestModel: [m1.id, m2.id],
            test_models.TestFKModel: [m3.id],
        }))


class LoadFetchedObjectsIntoContextsTest(SimpleTestCase):
    def test_none(self):
        context_loader.load_fetched_objects_into_contexts([], {}, {})

    def test_event_with_no_model_data(self):
        e = N(models.Event, persist_dependencies=False, context={'hi', 'hi'})
        context_loader.load_fetched_objects_into_contexts([e], {}, {})
        self.assertEquals(e, e)

    def test_one_event_w_model_data(self):
        m = N(test_models.TestModel, persist_dependencies=False, id=2)
        s = N(models.Source, persist_dependencies=False, id=1)
        hints = {
            s: {
                'key': {
                    'model_name': 'TestModel',
                    'app_name': 'tests',
                }
            }
        }
        e = N(models.Event, persist_dependencies=False, context={'key': m.id}, source=s)
        context_loader.load_fetched_objects_into_contexts([e], {test_models.TestModel: {m.id: m}}, hints)
        self.assertEquals(e.context, {'key': m})

    def test_one_event_w_list_model_data(self):
        m1 = N(test_models.TestModel, persist_dependencies=False, id=2)
        m2 = N(test_models.TestModel, persist_dependencies=False, id=3)
        s = N(models.Source, persist_dependencies=False, id=1)
        hints = {
            s: {
                'key': {
                    'model_name': 'TestModel',
                    'app_name': 'tests',
                }
            }
        }
        e = N(models.Event, persist_dependencies=False, context={'key': [m1.id, m2.id]}, source=s)
        context_loader.load_fetched_objects_into_contexts([e], {test_models.TestModel: {m1.id: m1, m2.id: m2}}, hints)
        self.assertEquals(e.context, {'key': [m1, m2]})


class TestLoadRenderersIntoEvents(SimpleTestCase):
    def test_no_mediums_or_renderers(self):
        events = [N(models.Event, context={}, persist_dependencies=False)]
        context_loader.load_renderers_into_events(events, [], [])
        self.assertEquals(events[0]._context_renderers, {})

    def test_mediums_and_no_renderers(self):
        events = [N(models.Event, context={}, persist_dependencies=False)]
        mediums = [N(models.Medium, persist_dependencies=False)]
        context_loader.load_renderers_into_events(events, mediums, [])
        self.assertEquals(events[0]._context_renderers, {})

    def test_mediums_w_renderers(self):
        s1 = N(models.Source, id=1, persist_dependencies=False)
        s2 = N(models.Source, id=2, persist_dependencies=False)
        e1 = N(models.Event, context={}, source=s1, persist_dependencies=False)
        e2 = N(models.Event, context={}, source=s2, persist_dependencies=False)
        rg1 = N(models.RenderingStyle, id=1, persist_dependencies=False)
        rg2 = N(models.RenderingStyle, id=2, persist_dependencies=False)
        m1 = N(models.Medium, id=1, rendering_style=rg1, persist_dependencies=False)
        m2 = N(models.Medium, id=2, rendering_style=rg2, persist_dependencies=False)
        cr1 = N(models.ContextRenderer, source=s1, rendering_style=rg1, id=1, persist_dependencies=False)
        cr2 = N(models.ContextRenderer, source=s2, rendering_style=rg1, id=2, persist_dependencies=False)
        cr3 = N(models.ContextRenderer, source=s1, rendering_style=rg2, id=3, persist_dependencies=False)

        context_loader.load_renderers_into_events([e1, e2], [m1, m2], [cr1, cr2, cr3])

        self.assertEquals(e1._context_renderers, {
            m1: cr1,
            m2: cr3,
        })
        self.assertEquals(e2._context_renderers, {
            m1: cr2
        })


class LoadContextsAndRenderersTest(TestCase):
    """
    Integration tests for loading contexts and renderers into events.
    """
    def test_none(self):
        context_loader.load_contexts_and_renderers([], [])

    def test_no_mediums(self):
        e = G(models.Event, context={})
        context_loader.load_contexts_and_renderers([e], [])
        self.assertEquals(e.context, {})

    def test_one_render_target_one_event(self):
        m1 = G(test_models.TestModel)
        s = G(models.Source)
        rg = G(models.RenderingStyle)
        e = G(models.Event, context={'key': m1.id}, source=s)
        medium = G(models.Medium, source=s, rendering_style=rg)
        G(models.ContextRenderer, rendering_style=rg, source=s, context_hints={
            'key': {
                'model_name': 'TestModel',
                'app_name': 'tests',
            }
        })

        context_loader.load_contexts_and_renderers([e], [medium])
        self.assertEquals(e.context, {'key': m1})

    def test_multiple_render_targets_multiple_events(self):
        test_m1 = G(test_models.TestModel)
        test_m2 = G(test_models.TestModel)
        test_m3 = G(test_models.TestModel)
        test_fk_m1 = G(test_models.TestFKModel)
        test_fk_m2 = G(test_models.TestFKModel)
        s1 = G(models.Source)
        s2 = G(models.Source)
        rg1 = G(models.RenderingStyle)
        rg2 = G(models.RenderingStyle)
        medium1 = G(models.Medium, source=s1, rendering_style=rg1)
        medium2 = G(models.Medium, source=s2, rendering_style=rg2)

        cr1 = G(models.ContextRenderer, rendering_style=rg1, source=s1, context_hints={
            'key': {
                'model_name': 'TestModel',
                'app_name': 'tests',
            }
        })
        cr2 = G(models.ContextRenderer, rendering_style=rg2, source=s2, context_hints={
            'key': {
                'model_name': 'TestModel',
                'app_name': 'tests',
            },
            'key2': {
                'model_name': 'TestFKModel',
                'app_name': 'tests',
            }
        })

        e1 = G(models.Event, context={'key': test_m1.id, 'key2': 'haha'}, source=s1)
        e2 = G(models.Event, context={'key': [test_m2.id, test_m3.id]}, source=s1)
        e3 = G(models.Event, context={'key2': test_fk_m1.id, 'key': test_m1.id}, source=s2)
        e4 = G(models.Event, context={'key2': test_fk_m2.id}, source=s2)

        context_loader.load_contexts_and_renderers([e1, e2, e3, e4], [medium1, medium2])
        self.assertEquals(e1.context, {'key': test_m1, 'key2': 'haha'})
        self.assertEquals(e2.context, {'key': [test_m2, test_m3]})
        self.assertEquals(e3.context, {'key2': test_fk_m1, 'key': test_m1})
        self.assertEquals(e4.context, {'key2': test_fk_m2})

        # Verify context renderers are put into the events properly
        self.assertEquals(e1._context_renderers, {
            medium1: cr1,
        })
        self.assertEquals(e2._context_renderers, {
            medium1: cr1,
        })
        self.assertEquals(e3._context_renderers, {
            medium2: cr2,
        })
        self.assertEquals(e4._context_renderers, {
            medium2: cr2,
        })

    def test_optimal_queries(self):
        fk1 = G(test_models.TestFKModel)
        fk11 = G(test_models.TestFKModel)
        fk2 = G(test_models.TestFKModel2)
        test_m1 = G(test_models.TestModel, fk=fk1, fk2=fk2)
        test_m1.fk_m2m.add(fk1, fk11)
        test_m2 = G(test_models.TestModel, fk=fk1, fk2=fk2)
        test_m2.fk_m2m.add(fk1, fk11)
        test_m3 = G(test_models.TestModel, fk=fk1, fk2=fk2)
        test_fk_m1 = G(test_models.TestFKModel)
        test_fk_m2 = G(test_models.TestFKModel)
        s1 = G(models.Source)
        s2 = G(models.Source)
        rg1 = G(models.RenderingStyle)
        rg2 = G(models.RenderingStyle)
        medium1 = G(models.Medium, source=s1, rendering_style=rg1)
        medium2 = G(models.Medium, source=s2, rendering_style=rg2)

        G(models.ContextRenderer, rendering_style=rg1, source=s1, context_hints={
            'key': {
                'model_name': 'TestModel',
                'app_name': 'tests',
                'select_related': ['fk'],
            }
        })
        G(models.ContextRenderer, rendering_style=rg2, source=s2, context_hints={
            'key': {
                'model_name': 'TestModel',
                'app_name': 'tests',
                'select_related': ['fk2'],
                'prefetch_related': ['fk_m2m'],
            },
            'key2': {
                'model_name': 'TestFKModel',
                'app_name': 'tests',
            }
        })

        e1 = G(models.Event, context={'key': test_m1.id, 'key2': 'haha'}, source=s1)
        e2 = G(models.Event, context={'key': [test_m2.id, test_m3.id]}, source=s1)
        e3 = G(models.Event, context={'key2': test_fk_m1.id, 'key': test_m1.id}, source=s2)
        e4 = G(models.Event, context={'key2': test_fk_m2.id}, source=s2)

        with self.assertNumQueries(6):
            context_loader.load_contexts_and_renderers([e1, e2, e3, e4], [medium1, medium2])
            self.assertEquals(e1.context['key'].fk, fk1)
            self.assertEquals(e2.context['key'][0].fk, fk1)
            self.assertEquals(e1.context['key'].fk2, fk2)
            self.assertEquals(e2.context['key'][0].fk2, fk2)
            self.assertEquals(set(e1.context['key'].fk_m2m.all()), set([fk1, fk11]))
            self.assertEquals(set(e2.context['key'][0].fk_m2m.all()), set([fk1, fk11]))
            self.assertEquals(e3.context['key'].fk, fk1)
